"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import statistics
import config as config
from scapy.all import *

RAW_DATA = config.PAYLOADDATA
SPLITBYTE = b'\x00'

class ExperimentsAnalyse :
    """ ExperimentAnalyse class
    This class is responsible of parsing the captured pcap files and analysing 
    the results of each experiments.

    Attributes
    ----------
    srcIP : str
        Provide the source IP

    dstIP : str
        Provide the destination IP
    
    sport : int
        Provide the source port
    
    dport : int
        Provide the destination port
    
    sndPkts : list
        Contain the sended packets, recorded from the source
    
    rcvPkts : list
        Contain the received packets, recorded from the destination
    
    exps : dict
        Contain the configuration of each experiment to analyse
    """

    def __init__(self, srcIP, dstIP, sport, dport, sndPkts, rcvPkts, exps):
        self.srcIP = srcIP
        self.dstIP = dstIP
        self.sport = sport
        self.dport = dport
        self.sndPkts = sndPkts
        self.rcvPkts = rcvPkts
        self.exps = exps

    """
    Compute the transmission delay for a set of packets
    """
    def compute_delay(self, sendIDs, rcvIDs):
        delays = []
        for sndPktID, sndTime in sendIDs.items():
            if sndPktID in rcvIDs:
                delay = float(rcvIDs[sndPktID] - sndTime) * 1000 # in ms
                delays.append(delay)
        
        if len(delays) < 1:
            return None
        else:
            return round(statistics.median(delays), 3)

    """
    Determine the state of an experiment
    """
    def determine_state(self, nbFilteredRcvPkts, nbNoFilteredRcvPkts):
        if nbFilteredRcvPkts > 0:
            return 'filtered'
        elif nbNoFilteredRcvPkts > 0:
            return 'forward'
        else:
            return 'drop'
    
    """
    Start the experiment analyse
    """
    def start_experiments_analyse(self):
        analyseResults = {}

        # Split UDP and TCP and other
        self.testsSndPktsUDP, self.testsSndPktsTCP, self.otherSndPkts = self.split_byProtocol(self.sndPkts)
        self.testsRcvPktsUDP, self.testsRcvPktsTCP, self.otherRcvPkts = self.split_byProtocol(self.rcvPkts)

        # Execute asked analysis
        for exp in self.exps:
            results = None
            if 'name' in exp.keys():
                if exp['name'] == 'StartTrace':
                    continue
                elif exp['name'] == 'EndTrace':
                    continue
                elif exp['name'] == 'NoHeader':
                    results = self.NoHeader_analyse(exp)
                elif exp['name'] == 'DestExtHdr':
                    results = self.DestExtHdr_analyse(exp)
                elif exp['name'] == 'HbHExtHdr':
                    results = self.HbHExtHdr_analyse(exp)
                elif exp['name'] == 'RoutingExtHdr':
                    results = self.RoutingExtHdr_analyse(exp)
                elif exp['name'] == 'FragExtHdr':
                    results = self.FragExtHdr_analyse(exp)
                elif exp['name'] == 'AHExtHdr':
                    results = self.AHExtHdr_analyse(exp)
                elif exp['name'] == 'ESPExtHdr':
                    results = self.ESPExtHdr_analyse(exp)
                elif exp['name'] == 'NoNextHdr':
                    results = self.NoNextHdr_analyse(exp)
                elif exp['name'] == 'Ethernet':
                    results = self.Ethernet_analyse(exp)
                else:
                    print("[ERROR]: Unknow analyse - {}".format(exp['name']))
                    continue
            else:
                print("[ERROR]: Exp has no name key")
                continue

            # Insert results
            for rslt in results:
                analyseResults[rslt['expName']] = rslt

        return analyseResults

    """
    Split a set of pkts according to its layer 4 procotols
    """
    def split_byProtocol(self, pkts):
        testPktsUDP = []
        testPktsTCP = []
        otherPkts = []

        # Filter by protocol
        for pkt in pkts:
            inserted = False
            if pkt.haslayer(IPv6) and pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                    if RAW_DATA in pkt[Raw].load:
                        if pkt.haslayer(UDP):
                            if pkt[UDP].sport == self.sport and pkt[UDP].dport == self.dport:
                                testPktsUDP.append(pkt)
                                inserted = True
                        elif pkt.haslayer(TCP):
                            if pkt[TCP].sport == self.sport and pkt[TCP].dport == self.dport and pkt[TCP].flags == 'S':
                                testPktsTCP.append(pkt)
                                inserted = True
                        else:
                            # Handle case of m=1 fragheader
                            if pkt.haslayer(IPv6ExtHdrFragment):
                                if pkt[IPv6ExtHdrFragment].nh == 17:
                                    testPktsUDP.append(pkt)
                                    inserted = True
                                if pkt[IPv6ExtHdrFragment].nh == 6:
                                    testPktsTCP.append(pkt)
                                    inserted = True

            if not inserted:
                otherPkts.append(pkt)

        return testPktsUDP, testPktsTCP, otherPkts

    """
    Analyse pkts related to NoHeader experiments
    """   
    def NoHeader_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return

        results = []
        # Iterate through protocols
        for proto in expConfig['proto']:
            validedSndpkts = []
            validedRcvpkts = []
            filteredRcvPkts = []
            sendIDs = {}
            recvIDs = {}

            # Configure variables
            if proto == 'UDP':
                testsSndPkts = self.testsSndPktsUDP
                testsRcvPkts = self.testsRcvPktsUDP
                IPv6NH = 17
            elif proto == 'TCP':
                testsSndPkts = self.testsSndPktsTCP
                testsRcvPkts = self.testsRcvPktsTCP
                IPv6NH = 6
            else:
                print('[ERROR] Unknow protocol')
                continue
                
            # Retrieve sended pkts
            for pkt in testsSndPkts:
                if pkt[IPv6].nh == IPv6NH:
                    sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                    validedSndpkts.append(pkt)

            # Retrieve received pkts
            for pkt in testsRcvPkts:
                recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                if recvID in sendIDs:
                    recvIDs[recvID] = pkt.time
                    if pkt[IPv6].nh == IPv6NH:
                        validedRcvpkts.append(pkt)
                    else:
                        filteredRcvPkts.append(pkt)

            # Determine state
            state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
            # Compute delay
            medianDelay = self.compute_delay(sendIDs, recvIDs)

            # Fill results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'state': state,
                'sndPktsCnt': len(validedSndpkts),
                'rcvPktsCnt': len(validedRcvpkts),
                'filteredPktsCnt': len(filteredRcvPkts),
                'medianDelay': medianDelay
            })
        
        return results
    
    """
    Analyse pkts related to destination header experiments
    """
    def DestExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys() or not 'types' in expConfig.keys():
            return

        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for action in type['action']:
                for other in type['other']:
                    for proto in expConfig['proto']:
                        validedSndpkts = []
                        validedRcvpkts = []
                        filteredRcvPkts = []
                        sendIDs = {}
                        recvIDs = {}

                        if action == 'skip':
                            if other == 'Unkw':
                                otype = 22
                            else:
                                otype = 1
                        elif action == 'discard':
                            if other == 'Unkw':
                                otype = 150
                            else:
                                otype = 129
                        else:
                            print('[ERROR]: Unknow action - {}'.format(action))
                            continue

                        # Configure variables
                        IPv6NH = 60
                        if proto == 'UDP':
                            testsSndPkts = self.testsSndPktsUDP
                            testsRcvPkts = self.testsRcvPktsUDP
                            IPv6ExtHdrDestOptNH = 17
                        elif proto == 'TCP':
                            testsSndPkts = self.testsSndPktsTCP
                            testsRcvPkts = self.testsRcvPktsTCP
                            IPv6ExtHdrDestOptNH = 6
                        else:
                            print('[ERROR] Unknow protocol')
                            continue

                        # Retrieve sended pkts
                        for pkt in testsSndPkts:
                            toInsert = False
                            if pkt[IPv6].nh == IPv6NH:
                                if pkt[IPv6ExtHdrDestOpt].nh == IPv6ExtHdrDestOptNH:
                                    if type['len'] <= 256:
                                        if len(pkt[IPv6ExtHdrDestOpt].options) == 1:
                                            if pkt[IPv6ExtHdrDestOpt].options[0].otype == otype and pkt[IPv6ExtHdrDestOpt].options[0].optlen == type['len'] - 4:
                                                toInsert = True
                                    else:
                                        if pkt[IPv6ExtHdrDestOpt].len == (type['len'] / 8) - 1:
                                            if len(pkt[IPv6ExtHdrDestOpt].options) > 1:
                                                toInsert = True
                            if toInsert:
                                sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                validedSndpkts.append(pkt)
                        
                        # Retrieve received pkts
                        for pkt in testsRcvPkts:
                            toInsert = False
                            recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                            if recvID in sendIDs:
                                recvIDs[recvID] = pkt.time
                                if pkt[IPv6].nh == IPv6NH:
                                    if pkt[IPv6ExtHdrDestOpt].nh == IPv6ExtHdrDestOptNH:
                                        if type['len'] <= 256:
                                            if len(pkt[IPv6ExtHdrDestOpt].options) == 1:
                                                if pkt[IPv6ExtHdrDestOpt].options[0].otype == otype and pkt[IPv6ExtHdrDestOpt].options[0].optlen == type['len'] - 4:
                                                    toInsert = True
                                        else:
                                            if pkt[IPv6ExtHdrDestOpt].len == (type['len'] / 8) - 1:
                                                if len(pkt[IPv6ExtHdrDestOpt].options) > 1:
                                                    toInsert = True
                                if toInsert:
                                    validedRcvpkts.append(pkt)
                                else:
                                    filteredRcvPkts.append(pkt)

                        # Determine state
                        state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
                        # Compute delay
                        medianDelay = self.compute_delay(sendIDs, recvIDs)

                        # Fill results
                        results.append({
                            'expName': "{}{}{}{}{}".format(expConfig['name'], str(type['len']), action, other if other != 'Unkw' else '', proto),
                            'state': state,
                            'sndPktsCnt': len(validedSndpkts),
                            'rcvPktsCnt': len(validedRcvpkts),
                            'filteredPktsCnt': len(filteredRcvPkts),
                            'medianDelay': medianDelay
                        })
        
        return results
    
    """
    Analyse pkts related to hop-by-hop header experiments
    """
    def HbHExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys() or not 'types' in expConfig.keys():
            return

        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for action in type['action']:
                for other in type['other']:
                    for proto in expConfig['proto']:
                        validedSndpkts = []
                        validedRcvpkts = []
                        filteredRcvPkts = []
                        sendIDs = {}
                        recvIDs = {}

                        if action == 'skip':
                            if other == 'Unkw':
                                otype = 54
                            else:
                                otype = 1
                        elif action == 'discard':
                            if other == 'Unkw':
                                otype = 182
                            else:
                                otype = 129
                        else:
                            print('[ERROR]: Unknow action - {}'.format(action))
                            continue

                        # Configure variables
                        IPv6NH = 0
                        if proto == 'UDP':
                            testsSndPkts = self.testsSndPktsUDP
                            testsRcvPkts = self.testsRcvPktsUDP
                            IPv6ExtHdrHbHOptNH = 17
                        elif proto == 'TCP':
                            testsSndPkts = self.testsSndPktsTCP
                            testsRcvPkts = self.testsRcvPktsTCP
                            IPv6ExtHdrHbHOptNH = 6
                        else:
                            print('[ERROR] Unknow protocol')
                            continue

                        # Retrieve sended pkts
                        for pkt in testsSndPkts:
                            toInsert = False
                            if pkt[IPv6].nh == IPv6NH:
                                if pkt[IPv6ExtHdrHopByHop].nh == IPv6ExtHdrHbHOptNH:
                                    if type['len'] <= 256:
                                        if len(pkt[IPv6ExtHdrHopByHop].options) == 1:
                                            if pkt[IPv6ExtHdrHopByHop].options[0].otype == otype and pkt[IPv6ExtHdrHopByHop].options[0].optlen == type['len'] - 4:
                                                toInsert = True
                                    else:
                                        if pkt[IPv6ExtHdrHopByHop].len == (type['len'] / 8) - 1:
                                            if len(pkt[IPv6ExtHdrHopByHop].options) > 1:
                                                toInsert = True
                            if toInsert:
                                sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                validedSndpkts.append(pkt)
                        
                        # Retrieve received pkts
                        for pkt in testsRcvPkts:
                            toInsert = False
                            recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                            if recvID in sendIDs:
                                recvIDs[recvID] = pkt.time
                                if pkt[IPv6].nh == IPv6NH:
                                    if pkt[IPv6ExtHdrHopByHop].nh == IPv6ExtHdrHbHOptNH:
                                        if type['len'] <= 256:
                                            if len(pkt[IPv6ExtHdrHopByHop].options) == 1:
                                                if pkt[IPv6ExtHdrHopByHop].options[0].otype == otype and pkt[IPv6ExtHdrHopByHop].options[0].optlen == type['len'] - 4:
                                                    toInsert = True
                                        else:
                                            if pkt[IPv6ExtHdrHopByHop].len == (type['len'] / 8) - 1:
                                                if len(pkt[IPv6ExtHdrHopByHop].options) > 1:
                                                    toInsert = True
                                if toInsert:
                                    validedRcvpkts.append(pkt)
                                else:
                                    filteredRcvPkts.append(pkt)

                        # Determine state
                        state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
                        # Compute delay
                        medianDelay = self.compute_delay(sendIDs, recvIDs)

                        results.append({
                            'expName': "{}{}{}{}{}".format(expConfig['name'], str(type['len']), action, other if other != 'Unkw' else '', proto),
                            'state': state,
                            'sndPktsCnt': len(validedSndpkts),
                            'rcvPktsCnt': len(validedRcvpkts),
                            'filteredPktsCnt': len(filteredRcvPkts),
                            'medianDelay': medianDelay
                        })
    
        return results

    """
    Analyse pkts related to routing header experiments
    """
    def RoutingExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys() or not 'types' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for proto in expConfig['proto']:
                validedSndpkts = []
                validedRcvpkts = []
                filteredRcvPkts = []
                sendIDs = {}
                recvIDs = {}

                # Configure variables
                IPv6NH = 43
                if proto == 'UDP':
                    testsSndPkts = self.testsSndPktsUDP
                    testsRcvPkts = self.testsRcvPktsUDP
                    IPv6ExtHdrRoutingNH = 17
                elif proto == 'TCP':
                    testsSndPkts = self.testsSndPktsTCP
                    testsRcvPkts = self.testsRcvPktsTCP
                    IPv6ExtHdrRoutingNH = 6
                else:
                    print('[ERROR] Unknow protocol')
                    continue
                    
                # Retrieve sended pkts
                for pkt in testsSndPkts:
                    toInsert = False
                    if pkt[IPv6].nh == IPv6NH:
                        if type == 4:
                            if pkt.haslayer(IPv6ExtHdrSegmentRouting):
                                if pkt[IPv6ExtHdrSegmentRouting].nh == IPv6ExtHdrRoutingNH:
                                    if pkt[IPv6ExtHdrSegmentRouting].type == type:
                                        if pkt[IPv6ExtHdrSegmentRouting].segleft == 0: 
                                            if pkt[IPv6ExtHdrSegmentRouting].addresses == [self.dstIP]:
                                                toInsert = True
                        else:
                            if pkt.haslayer(IPv6ExtHdrRouting):
                                if pkt[IPv6ExtHdrRouting].nh == IPv6ExtHdrRoutingNH:
                                    segleft = 0
                                    if type == 2:
                                        segleft = 1
                                    if pkt[IPv6ExtHdrRouting].segleft == segleft:
                                        if pkt[IPv6ExtHdrRouting].type == type:
                                            if type == 0 or type == 1 or type == 2 or type == 3:
                                                if pkt[IPv6ExtHdrRouting].addresses == [self.dstIP]:
                                                    toInsert = True
                                            elif type == 5:
                                                if pkt[IPv6ExtHdrRouting].reserved == 0x00080000:
                                                    toInsert = True
                                            elif type == 6:
                                                if pkt[IPv6ExtHdrRouting].reserved == 0x00000008:
                                                    toInsert = True
                    if toInsert:
                        sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                        validedSndpkts.append(pkt)
    
                # Retrieve received pkts
                for pkt in testsRcvPkts:
                    toInsert = False
                    recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                    if recvID in sendIDs:
                        recvIDs[recvID] = pkt.time
                        if pkt[IPv6].nh == IPv6NH:
                            if type == 4:
                                if pkt.haslayer(IPv6ExtHdrSegmentRouting):
                                    if pkt[IPv6ExtHdrSegmentRouting].nh == IPv6ExtHdrRoutingNH:
                                        if pkt[IPv6ExtHdrSegmentRouting].type == type:
                                            if pkt[IPv6ExtHdrSegmentRouting].segleft == 0: 
                                                if pkt[IPv6ExtHdrSegmentRouting].addresses == [self.dstIP]:
                                                    toInsert = True
                            else:
                                if pkt.haslayer(IPv6ExtHdrRouting):
                                    if pkt[IPv6ExtHdrRouting].nh == IPv6ExtHdrRoutingNH:
                                        segleft = 0
                                        if type == 2:
                                            segleft = 1
                                        if pkt[IPv6ExtHdrRouting].segleft == segleft:
                                            if pkt[IPv6ExtHdrRouting].type == type:
                                                if type == 0 or type == 1 or type == 2 or type == 3:
                                                    if pkt[IPv6ExtHdrRouting].addresses == [self.dstIP]:
                                                        toInsert = True
                                                elif type == 5:
                                                    if pkt[IPv6ExtHdrRouting].reserved == 0x00080000:
                                                        toInsert = True
                                                elif type == 6:
                                                    if pkt[IPv6ExtHdrRouting].reserved == 0x00000008:
                                                        toInsert = True
                        if toInsert:
                            validedRcvpkts.append(pkt)
                        else:
                            filteredRcvPkts.append(pkt)
                
                # Determine state
                state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
                # Compute delay
                medianDelay = self.compute_delay(sendIDs, recvIDs)

                # Fill results
                results.append({
                    'expName': "{}Type{}{}".format(expConfig['name'], type, proto),
                    'state': state,
                    'sndPktsCnt': len(validedSndpkts),
                    'rcvPktsCnt': len(validedRcvpkts),
                    'filteredPktsCnt': len(filteredRcvPkts),
                    'medianDelay': medianDelay
                })
            
        return results
    
    """
    Analyse pkts related to fragmentation header experiments
    """
    def FragExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys() or not 'types' in expConfig.keys() or not 'm' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for proto in expConfig['proto']:
                for m in expConfig['m']:
                    validedSndpkts = []
                    validedRcvpkts = []
                    filteredRcvPkts = []
                    sendIDs = {}
                    recvIDs = {}
                    
                    # Configure variables
                    IPv6NH = 44
                    if proto == 'UDP':
                        testsSndPkts = self.testsSndPktsUDP
                        testsRcvPkts = self.testsRcvPktsUDP
                        IPv6ExtHdrFragmentNH = 17
                    elif proto == 'TCP':
                        testsSndPkts = self.testsSndPktsTCP
                        testsRcvPkts = self.testsRcvPktsTCP
                        IPv6ExtHdrFragmentNH = 6
                    else:
                        print('[ERROR] Unknow protocol')
                        continue
                        
                    # Retrieve sended pkts
                    for pkt in testsSndPkts:
                        if pkt[IPv6].nh == IPv6NH and pkt[IPv6].plen == int(type) - 54:
                            if pkt[IPv6ExtHdrFragment].nh == IPv6ExtHdrFragmentNH and pkt[IPv6ExtHdrFragment].m == m:
                                sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                validedSndpkts.append(pkt)
                
                    # Retrieve received pkts
                    for pkt in testsRcvPkts:
                        toInsert = False
                        recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                        if recvID in sendIDs:
                            recvIDs[recvID] = pkt.time
                            if pkt[IPv6].nh == IPv6NH and pkt[IPv6].plen == int(type) - 54:
                                if pkt[IPv6ExtHdrFragment].nh == IPv6ExtHdrFragmentNH and pkt[IPv6ExtHdrFragment].m == m:
                                    toInsert = True
                            
                            if toInsert:
                                validedRcvpkts.append(pkt)
                            else:
                                filteredRcvPkts.append(pkt)
                
                    # Determine state
                    state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
                    # Compute delay
                    medianDelay = self.compute_delay(sendIDs, recvIDs)

                    # Fill results
                    results.append({
                        'expName': "{}{}m{}{}".format(expConfig['name'], type, m, proto),
                        'state': state,
                        'sndPktsCnt': len(validedSndpkts),
                        'rcvPktsCnt': len(validedRcvpkts),
                        'filteredPktsCnt': len(filteredRcvPkts),
                        'medianDelay': medianDelay
                    })
        
        return results
    
    """
    Analyse pkts related to authentication header experiments
    """
    def AHExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedSndpkts = []
            validedRcvpkts = []
            filteredRcvPkts = []
            sendIDs = {}
            recvIDs = {}

            # Configure variables
            testsSndPkts = self.otherSndPkts
            testsRcvPkts = self.otherRcvPkts
            IPv6NH = 51
            if proto == 'UDP':
                AHNH = 17
            elif proto == 'TCP':
                AHNH = 6
            else:
                print('[ERROR] Unknow protocol')
                continue
                
            # Retrieve sended pkts
            for pkt in testsSndPkts:
                if pkt.haslayer(IPv6) and pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                    if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                        if RAW_DATA in pkt[Raw].load:
                            if pkt[IPv6].nh == IPv6NH and pkt.haslayer(AH):
                                if pkt[AH].nh == AHNH:
                                    if struct.pack(">H", self.sport) == pkt[Raw].load[0:2] and struct.pack(">H", self.dport) == pkt[Raw].load[2:4]:
                                        sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                        validedSndpkts.append(pkt)
                    
            # Retrieve received pkts
            for pkt in testsRcvPkts:
                toInsert = False
                if pkt.haslayer(IPv6) and pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                    if RAW_DATA in pkt[Raw].load:
                        recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                        if recvID in sendIDs:
                            recvIDs[recvID] = pkt.time
                            if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                                if pkt[IPv6].nh == IPv6NH and pkt.haslayer(AH):
                                    if pkt[AH].nh == AHNH:
                                        if struct.pack(">H", self.sport) == pkt[Raw].load[0:2] and struct.pack(">H", self.dport) == pkt[Raw].load[2:4]:
                                            toInsert = True
                            if toInsert:
                                validedRcvpkts.append(pkt)
                            else:
                                filteredRcvPkts.append(pkt)
          
            # Determine state
            state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
            # Compute delay
            medianDelay = self.compute_delay(sendIDs, recvIDs)

            # Fill results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'state': state,
                'sndPktsCnt': len(validedSndpkts),
                'rcvPktsCnt': len(validedRcvpkts),
                'filteredPktsCnt': len(filteredRcvPkts),
                'medianDelay': medianDelay
            })
        
        return results
    
    """
    Analyse pkts related to ESP header experiments
    """
    def ESPExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedSndpkts = []
            validedRcvpkts = []
            filteredRcvPkts = []
            sendIDs = {}
            recvIDs = {}

            # Configure variables
            IPv6NH = 50
            if proto == 'UDP' or proto == 'TCP':
                testsSndPkts = self.otherSndPkts
                testsRcvPkts = self.otherRcvPkts
            else:
                print('[ERROR] Unknow protocol')
                continue
                
            # Retrieve sended pkts
            for pkt in testsSndPkts:
                if pkt.haslayer(IPv6) and not pkt.haslayer(IPerror6):
                    if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                        if pkt[IPv6].nh == IPv6NH and pkt.haslayer(ESP):
                            sendIDs[int(pkt[ESP].seq)] = pkt.time
                            validedSndpkts.append(pkt)
                    
            # Retrieve received pkts
            for pkt in testsRcvPkts:
                toInsert = False
                if pkt.haslayer(IPv6) and not pkt.haslayer(IPerror6):
                    if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                        if pkt[IPv6].nh == IPv6NH and pkt.haslayer(ESP):
                            recvID = int(pkt[ESP].seq)
                            if recvID in sendIDs:
                                recvIDs[recvID] = pkt.time
                                toInsert = True
                                
                            if toInsert:
                                validedRcvpkts.append(pkt)
                            else:
                                filteredRcvPkts.append(pkt)
          
            # Determine state
            state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
            # Compute delay
            medianDelay = self.compute_delay(sendIDs, recvIDs)

            # Fill results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'state': state,
                'sndPktsCnt': len(validedSndpkts),
                'rcvPktsCnt': len(validedRcvpkts),
                'filteredPktsCnt': len(filteredRcvPkts),
                'medianDelay': medianDelay
            })
        
        return results

    """
    Analyse pkts related to no next header experiments
    """
    def NoNextHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedSndpkts = []
            validedRcvpkts = []
            filteredRcvPkts = []
            sendIDs = {}
            recvIDs = {}

            # Configure variables
            testsSndPkts = self.otherSndPkts
            testsRcvPkts = self.otherRcvPkts
            if proto == 'noL4':
                IPv6NH = 59
            else:
                print('[ERROR] Unknow protocol')
                continue
                
            # Retrieve sended pkts
            for pkt in testsSndPkts:
                if pkt.haslayer(IPv6) and pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                    if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                        if pkt [IPv6].nh == IPv6NH:
                            if RAW_DATA in pkt[Raw].load:
                                sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                validedSndpkts.append(pkt)
                    
            # Retrieve received pkts
            for pkt in testsRcvPkts:
                toInsert = False
                if pkt.haslayer(Raw):
                    if RAW_DATA in pkt[Raw].load and not pkt.haslayer(IPerror6):
                        recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                        if recvID in sendIDs:
                            recvIDs[recvID] = pkt.time
                            if pkt.haslayer(IPv6) and pkt.haslayer(Raw):
                                if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                                    if pkt [IPv6].nh == IPv6NH:
                                        toInsert = True
                            if toInsert:
                                validedRcvpkts.append(pkt)
                            else:
                                filteredRcvPkts.append(pkt)
          
            # Determine state
            state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
            # Compute delay
            medianDelay = self.compute_delay(sendIDs, recvIDs)

            # Fill results
            results.append({
                'expName': "{}".format(expConfig['name']),
                'state': state,
                'sndPktsCnt': len(validedSndpkts),
                'rcvPktsCnt': len(validedRcvpkts),
                'filteredPktsCnt': len(filteredRcvPkts),
                'medianDelay': medianDelay
            })
        
        return results
    
    """
    Analyse pkts related to ethernet header experiments
    """
    def Ethernet_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedSndpkts = []
            validedRcvpkts = []
            filteredRcvPkts = []
            sendIDs = {}
            recvIDs = {}

            # Configure variables
            testsSndPkts = self.otherSndPkts
            testsRcvPkts = self.otherRcvPkts
            if proto == 'noL4':
                IPv6NH = 143
            else:
                print('[ERROR] Unknow protocol')
                continue
                
            # Retrieve sended pkts
            for pkt in testsSndPkts:
                if pkt.haslayer(IPv6) and pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                    if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                        if pkt[IPv6].nh == IPv6NH:
                            if RAW_DATA in pkt[Raw].load:
                                sendIDs[int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])] = pkt.time
                                validedSndpkts.append(pkt)
                    
            # Retrieve received pkts
            for pkt in testsRcvPkts:
                toInsert = False
                if pkt.haslayer(Raw) and not pkt.haslayer(IPerror6):
                    if RAW_DATA in pkt[Raw].load:
                        recvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                        if recvID in sendIDs:
                            recvIDs[recvID] = pkt.time
                            if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and pkt[IPv6].hlim > 200:
                                if pkt [IPv6].nh == IPv6NH:
                                    toInsert = True
                            if toInsert:
                                validedRcvpkts.append(pkt)
                            else:
                                filteredRcvPkts.append(pkt)
          
            # Determine state
            state = self.determine_state(len(filteredRcvPkts), len(validedRcvpkts))
            # Compute delay
            medianDelay = self.compute_delay(sendIDs, recvIDs)

            # Fill results
            results.append({
                'expName': "{}".format(expConfig['name']),
                'state': state,
                'sndPktsCnt': len(validedSndpkts),
                'rcvPktsCnt': len(validedRcvpkts),
                'filteredPktsCnt': len(filteredRcvPkts),
                'medianDelay': medianDelay
            })

        return results