"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import config as config
from utils import *
from scapy.all import *

MAXHLIM = 100
DELTATIME = 1.5
TRACEINFOSFILENAME = "traceroutes_infos.txt"
RAW_DATA = config.PAYLOADDATA
SPLITBYTE = b'\x00'

class TracerouteAnalyse:
    """ TracerouteAnalyse class
    This class is responsible of parsing the captured pcap files and analysing 
    the traceroute performed for each experiments.

    Attributes
    ----------
    srcIP : str
        Provide the source IP

    dstIP : str
        Provide the destination IP
    
    sport : int
        Provide the source port
    
    dport : int
        Provide the destination port
    
    tracePkts : list
        Contain the sended and received packets, recorded from the source
    
    exps : dict
        Contain the configuration of each experiment to analyse
    
    traceInfosPath : str
        Path to the file containing the timestamp of each traceroute performed
        during the probing phase
    """

    def __init__(self, srcIP, dstIP, sport, dport, tracePkts, exps, traceInfosPath):
        self.srcIP = srcIP
        self.dstIP = dstIP
        self.sport = sport
        self.dport = dport
        self.tracePkts = tracePkts
        self.exps = exps
        self.traceInfosPath = traceInfosPath
    
    """
    Start the traceroute analyse
    """
    def start_traceroutes_analyse(self):
        analyseResults = {}

        # Split traceroute probes by UDP and TCP
        self.traceProbesUDP, self.traceProbesTCP, self.otherPkts = self.split_probes_byProtocol(self.tracePkts)
        # Split responses (ICMP and SYN/ACK) by UDP and TCP
        self.respUDP, self.respTCP, self.otherPkts = self.split_responses_byProtocol(self.otherPkts)

        # Execute asked analysis
        for exp in self.exps:
            results = None
            if 'name' in exp.keys():
                if exp['name'] == 'StartTrace':
                    results = self.NoHeader_analyse(exp)
                elif exp['name'] == 'EndTrace':
                    results = self.NoHeader_analyse(exp)
                elif exp['name'] == 'NoHeader':
                    results = self.NoHeader_analyse(exp)
                elif exp['name'] == 'DestExtHdr':
                    results = self.DestExtHdr_analyse(exp)
                elif exp['name'] == 'HbHExtHdr':
                    results = self.HbHExtHdr_analyse(exp)
                elif exp['name'] == 'RoutingExtHdr':
                    results = self.RoutingExtHdr_analyse(exp)
                elif exp['name'] == 'FragExtHdr':
                    results = self.FragExtHdr_analyse(exp)
                elif exp['name'] == 'AHExtHdr':
                    results = self.AHExtHdr_analyse(exp)
                elif exp['name'] == 'ESPExtHdr':
                    results = self.ESPExtHdr_analyse(exp)
                elif exp['name'] == 'NoNextHdr':
                    results = self.NoNextHdr_analyse(exp)
                elif exp['name'] == 'Ethernet':
                    results = self.Ethernet_analyse(exp)
                else:
                    print("[ERROR]: Unknow analyse - {}".format(exp['name']))
                    continue
            else:
                print("[ERROR]: Exp has no name key")
                continue
            
            # Format results 
            for rslt in results:
                if rslt is None:
                    continue
                    
                hops = {}
                maxHlimSeen = None
                
                # Insert hops
                lastIPSeen = None
                isHostReach = False
                for hlim in range(1, rslt['nbProbesSnd']+1):
                    if hlim in rslt['hops']:
                        hopInfos = rslt['hops'][hlim]
                        hopIP = hopInfos[0][IPv6].src

                        # Get response type
                        if hopInfos[0].haslayer(ICMPv6TimeExceeded):
                            hopMsgType = "TimeExceeded"
                        elif hopInfos[0].haslayer(ICMPv6ParamProblem):
                            hopMsgType = "ParamProblem"
                        elif hopInfos[0].haslayer(ICMPv6DestUnreach):
                            hopMsgType = "DestUnreach"
                        elif hopInfos[0].haslayer(TCP):
                            hopMsgType = "SYN/ACK"
                        else:
                            hopMsgType = "Other"

                        # Filter to get only valid hops
                        if lastIPSeen != hopIP:
                            # Insert hop
                            hops[hlim] = {
                                'ip': hopIP,
                                'rtt': hopInfos[1],
                                'isFilter': hopInfos[2],
                                'respType': hopMsgType,
                            }

                            lastIPSeen = hopIP

                            # Stop if destination reached
                            if hopIP == self.dstIP:
                                isHostReach = True
                                maxHlimSeen = hlim
                                break
                        else:
                            # Handle case of multiple ParamProblem from same IP
                            break

                if not isHostReach:
                    maxHlimSeen = rslt['nbProbesSnd']
                
                analyseResults[rslt['expName']] = {
                    'maxHlimSeen': maxHlimSeen,
                    'hops': hops
                }

        return analyseResults
    
    """
    Extract probes and split them according to their layer 4 protocol
    """
    def split_probes_byProtocol(self, tracePkts):
        traceProbesUDP = []
        traceProbesTCP = []
        otherPkts = []

        # Filter by protocol
        for pkt in tracePkts:
            inserted = False
            if pkt.haslayer(IPv6) and pkt.haslayer(Raw):
                if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and 1 <= pkt[IPv6].hlim <= MAXHLIM:
                    if not (pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach)):
                        if RAW_DATA in pkt[Raw].load:
                            if pkt.haslayer(UDP):
                                if pkt[UDP].sport == self.sport and pkt[UDP].dport == self.dport:
                                    traceProbesUDP.append(pkt)
                                    inserted = True
                            elif pkt.haslayer(TCP):
                                if pkt[TCP].sport == self.sport and pkt[TCP].dport == self.dport:
                                    traceProbesTCP.append(pkt)
                                    inserted = True
                            else:
                                # Handle case of m=1 fragheader
                                if pkt.haslayer(IPv6ExtHdrFragment):
                                    if pkt[IPv6ExtHdrFragment].nh == 17:
                                        traceProbesUDP.append(pkt)
                                        inserted = True
                                    if pkt[IPv6ExtHdrFragment].nh == 6:
                                        traceProbesTCP.append(pkt)
                                        inserted = True
        
            if not inserted:
                otherPkts.append(pkt)

        return traceProbesUDP, traceProbesTCP, otherPkts

    """
    Extract resposnes and split them according to their type
    """
    def split_responses_byProtocol(self, pkts):
        respUDP = []
        respTCP = []
        otherPkts = []

        for pkt in pkts:
            inserted = False
            if pkt.haslayer(IPv6):
                if pkt[IPv6].dst == self.srcIP:
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        if pkt.haslayer(Raw):
                            if RAW_DATA in pkt[Raw].load:
                                if pkt.haslayer(UDP):
                                    if pkt[UDP].sport == self.sport and pkt[UDP].dport == self.dport:
                                        respUDP.append(pkt)
                                        inserted = True
                                elif pkt.haslayer(UDPerror):
                                    if pkt[UDPerror].sport == self.sport and pkt[UDPerror].dport == self.dport:
                                        respUDP.append(pkt)
                                        inserted = True
                                elif pkt.haslayer(TCP):
                                    if pkt[TCP].sport == self.sport and pkt[TCP].dport == self.dport:
                                        respTCP.append(pkt)
                                        inserted = True
                                elif pkt.haslayer(TCPerror):
                                    if pkt[TCPerror].sport == self.sport and pkt[TCPerror].dport == self.dport:
                                        respTCP.append(pkt)
                                        inserted = True
                                else:
                                    # Handle case of m=1 frag header
                                    if pkt.haslayer(IPv6ExtHdrFragment):
                                        if pkt[IPv6ExtHdrFragment].nh == 17:
                                            respUDP.append(pkt)
                                            inserted = True
                                        if pkt[IPv6ExtHdrFragment].nh == 6:
                                            respTCP.append(pkt)
                                            inserted = True
                    # Handle SYN/ACK response                        
                    elif pkt.haslayer(TCP):
                        if pkt[TCP].sport == self.dport and pkt[TCP].dport == self.sport:
                                if pkt[TCP].flags == 'SA':
                                    respTCP.append(pkt)
                                    inserted = True
            
            if not inserted:
                otherPkts.append(pkt)

        return respUDP, respTCP, otherPkts

    """
    Match probes and responses according to the recorded IDs
    """
    def match_probes_responses(self, probes, validICMPs, filteredICMPs, synAcks):
        hops = {}
        maxHlimSeen = 0

        # Iterate over the probes
        for probe in probes:
            resp = None

            # Retrieve probe ID and hlim
            if probe.haslayer(ESP):
                sendID = int(probe[ESP].seq)
            else:
                sendID = int(probe[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])

            probeHlim = probe[IPv6].hlim

            # Record the highest hlim
            if probeHlim >= maxHlimSeen:
                maxHlimSeen = probeHlim

            # Find corresponding responses
            if sendID in validICMPs:
                resp = validICMPs[sendID]
                isFiltered = False
            elif sendID in synAcks:
                resp = synAcks[sendID]
                isFiltered = False
            elif sendID in filteredICMPs:
                resp = filteredICMPs[sendID]
                isFiltered = True
            else:
                continue

            # Compute rtt (in ms)
            rtt = round(float(resp.time - probe.time) * 1000, 3) # in ms
            
            # Check if hops already identified and insert hop
            if probeHlim in hops:
                if rtt <= hops[probeHlim][1]:
                    hops[probeHlim] = (resp, rtt, isFiltered)
            else:
                hops[probeHlim] = (resp, rtt, isFiltered)
        
        return hops, maxHlimSeen

    """
    Analyse pkts related to NoHeader experiments
    """
    def NoHeader_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedProbes = []
            validedICMPResps = {}
            filteredICMPResps = {}
            validedSAResps = {}

            # Configure variables
            if proto == 'UDP':
                traceProbes = self.traceProbesUDP
                respPkts = self.respUDP
                IPv6NH = 17
            elif proto == 'TCP':
                traceProbes = self.traceProbesTCP
                respPkts = self.respTCP
                IPv6NH = 6
            else:
                print('[ERROR] Unknow protocol')
                continue

            # Retrieve traceroute start and end time
            traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expConfig['name'])
            if traceInfos is None:
                return None
            
            startTime = float(traceInfos['START_TIME'])
            endTime = float(traceInfos['END_TIME'])

            # Retrieve probes
            for pkt in traceProbes:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt[IPv6].nh == IPv6NH:
                        validedProbes.append(pkt)
            
            # Retrieve responses
            for pkt in respPkts:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                        if pkt[IPerror6].nh == IPv6NH:
                            validedICMPResps[rcvID] = pkt
                        else:
                            filteredICMPResps[rcvID] = pkt

                    elif pkt.haslayer(TCP):
                        if pkt[TCP].flags == 'SA':
                            rcvID = pkt[TCP].ack - 1
                            validedSAResps[rcvID] = pkt
                
            # Match probe with corresponding response
            hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
            
            # Record results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'nbProbesSnd': maxHlimSeen,
                'hops': hops, 
            })
            
        return results
    
    """
    Analyse pkts related to destination header experiments
    """
    def DestExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for action in type['action']:
                for other in type['other']:
                    for proto in expConfig['proto']:
                        validedProbes = []
                        validedICMPResps = {}
                        filteredICMPResps = {}
                        validedSAResps = {}

                        # Define otype
                        if action == 'skip':
                            if other == 'Unkw':
                                otype = 22
                            else:
                                otype = 1
                        elif action == 'discard':
                            if other == 'Unkw':
                                otype = 150
                            else:
                                otype = 129
                        else:
                            print('[ERROR]: Unknow action - {}'.format(action))
                            continue

                        # Configure variables
                        IPv6NH = 60
                        if proto == 'UDP':
                            traceProbes = self.traceProbesUDP
                            respPkts = self.respUDP
                            IPv6ExtHdrDestOptNH = 17
                        elif proto == 'TCP':
                            traceProbes = self.traceProbesTCP
                            respPkts = self.respTCP
                            IPv6ExtHdrDestOptNH = 6
                        else:
                            print('[ERROR] Unknow protocol')
                            continue
                        
                        # Compute full expName
                        expName = "Dest" + str(type['len']) + action
                        if other != 'Unkw':
                            expName += other
                        
                        # Retrieve traceroute start and end time
                        traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
                        if traceInfos is None:
                            return None
                        
                        startTime = float(traceInfos['START_TIME'])
                        endTime = float(traceInfos['END_TIME'])

                        # Retrieve probes
                        for pkt in traceProbes:
                            toInsert = False
                            if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                                if pkt[IPv6].nh == IPv6NH:
                                    if pkt[IPv6ExtHdrDestOpt].nh == IPv6ExtHdrDestOptNH:
                                        if type['len'] <= 256:
                                            if len(pkt[IPv6ExtHdrDestOpt].options) == 1:
                                                if pkt[IPv6ExtHdrDestOpt].options[0].otype == otype and pkt[IPv6ExtHdrDestOpt].options[0].optlen == type['len'] - 4:
                                                    toInsert = True
                                        else:
                                            if pkt[IPv6ExtHdrDestOpt].len == (type['len'] / 8) - 1:
                                                if len(pkt[IPv6ExtHdrDestOpt].options) > 1:
                                                    toInsert = True
                                if toInsert:
                                    validedProbes.append(pkt)
                        
                        # Retrieve responses
                        for pkt in respPkts:
                            if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                                if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                                    rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                    toInsert = False
                                    if pkt[IPerror6].nh == IPv6NH:
                                        if pkt[IPv6ExtHdrDestOpt].nh == IPv6ExtHdrDestOptNH:
                                            if type['len'] <= 256:
                                                if len(pkt[IPv6ExtHdrDestOpt].options) == 1:
                                                    if pkt[IPv6ExtHdrDestOpt].options[0].otype == otype and pkt[IPv6ExtHdrDestOpt].options[0].optlen == type['len'] - 4:
                                                        toInsert = True
                                            else:
                                                if pkt[IPv6ExtHdrDestOpt].len == (type['len'] / 8) - 1:
                                                    if len(pkt[IPv6ExtHdrDestOpt].options) > 1:
                                                        toInsert = True
                                    if toInsert:
                                        validedICMPResps[rcvID] = pkt
                                    else:
                                        filteredICMPResps[rcvID] = pkt

                                elif pkt.haslayer(TCP):
                                    if pkt[TCP].flags == 'SA':
                                        rcvID = pkt[TCP].ack - 1
                                        validedSAResps[rcvID] = pkt
                            
                        # Match probe with corresponding response
                        hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)

                        # Record results
                        results.append({
                            'expName': "{}{}{}{}{}".format(expConfig['name'], str(type['len']), action, other if other != 'Unkw' else '', proto),
                            'nbProbesSnd': maxHlimSeen,
                            'hops': hops, 
                        })
                        
        return results
    
    """
    Analyse pkts related to hop-by-hop header experiments
    """
    def HbHExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for action in type['action']:
                for other in type['other']:
                    for proto in expConfig['proto']:
                        validedProbes = []
                        validedICMPResps = {}
                        filteredICMPResps = {}
                        validedSAResps = {}

                        # Define otype
                        if action == 'skip':
                            if other == 'Unkw':
                                otype = 54
                            else:
                                otype = 1
                        elif action == 'discard':
                            if other == 'Unkw':
                                otype = 182
                            else:
                                otype = 129
                        else:
                            print('[ERROR]: Unknow action - {}'.format(action))
                            continue

                        # Configure variables
                        IPv6NH = 0
                        if proto == 'UDP':
                            traceProbes = self.traceProbesUDP
                            respPkts = self.respUDP
                            IPv6ExtHdrHbHOptNH = 17
                        elif proto == 'TCP':
                            traceProbes = self.traceProbesTCP
                            respPkts = self.respTCP
                            IPv6ExtHdrHbHOptNH = 6
                        else:
                            print('[ERROR] Unknow protocol')
                            continue
                        
                        # Compute full expName
                        expName = "HbH" + str(type['len']) + action
                        if other != 'Unkw':
                            expName += other

                        # Retrieve traceroute start and end time
                        traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
                        if traceInfos is None:
                            return None
                        
                        startTime = float(traceInfos['START_TIME'])
                        endTime = float(traceInfos['END_TIME'])

                        # Retrieve probes
                        for pkt in traceProbes:
                            toInsert = False
                            if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                                if pkt[IPv6].nh == IPv6NH:
                                    if pkt[IPv6ExtHdrHopByHop].nh == IPv6ExtHdrHbHOptNH:
                                        if type['len'] <= 256:
                                            if len(pkt[IPv6ExtHdrHopByHop].options) == 1:
                                                if pkt[IPv6ExtHdrHopByHop].options[0].otype == otype and pkt[IPv6ExtHdrHopByHop].options[0].optlen == type['len'] - 4:
                                                    toInsert = True
                                        else:
                                            if pkt[IPv6ExtHdrHopByHop].len == (type['len'] / 8) - 1:
                                                if len(pkt[IPv6ExtHdrHopByHop].options) > 1:
                                                    toInsert = True
                                if toInsert:
                                    validedProbes.append(pkt)
                        
                        # Retrieve responses
                        for pkt in respPkts:
                            if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                                if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                                    rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                    toInsert = False
                                    if pkt[IPerror6].nh == IPv6NH:
                                        if pkt[IPv6ExtHdrHopByHop].nh == IPv6ExtHdrHbHOptNH:
                                            if type['len'] <= 256:
                                                if len(pkt[IPv6ExtHdrHopByHop].options) == 1:
                                                    if pkt[IPv6ExtHdrHopByHop].options[0].otype == otype and pkt[IPv6ExtHdrHopByHop].options[0].optlen == type['len'] - 4:
                                                        toInsert = True
                                            else:
                                                if pkt[IPv6ExtHdrHopByHop].len == (type['len'] / 8) - 1:
                                                    if len(pkt[IPv6ExtHdrHopByHop].options) > 1:
                                                        toInsert = True
                                    if toInsert:
                                        validedICMPResps[rcvID] = pkt
                                    else:
                                        filteredICMPResps[rcvID] = pkt

                                elif pkt.haslayer(TCP):
                                    if pkt[TCP].flags == 'SA':
                                        rcvID = pkt[TCP].ack - 1
                                        validedSAResps[rcvID] = pkt
                            
                        # Match probe with corresponding response
                        hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
                        
                        # Record results
                        results.append({
                            'expName': "{}{}{}{}{}".format(expConfig['name'], str(type['len']), action, other if other != 'Unkw' else '', proto),
                            'nbProbesSnd': maxHlimSeen,
                            'hops': hops, 
                        })
                        
        return results
    
    """
    Analyse pkts related to routing header experiments
    """
    def RoutingExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for proto in expConfig['proto']:
                validedProbes = []
                validedICMPResps = {}
                filteredICMPResps = {}
                validedSAResps = {}

                # Configure variables
                IPv6NH = 43
                if proto == 'UDP':
                    traceProbes = self.traceProbesUDP
                    respPkts = self.respUDP
                    IPv6ExtHdrRoutingNH = 17
                elif proto == 'TCP':
                    traceProbes = self.traceProbesTCP
                    respPkts = self.respTCP
                    IPv6ExtHdrRoutingNH = 6
                else:
                    print('[ERROR] Unknow protocol')
                    continue
                
                # Compute full expName
                expName = "RoutingType" + str(type)

                # Retrieve traceroute start and end time
                traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
                if traceInfos is None:
                    return None
                
                startTime = float(traceInfos['START_TIME'])
                endTime = float(traceInfos['END_TIME'])

                # Retrieve probes
                for pkt in traceProbes:
                    toInsert = False
                    if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                        if pkt[IPv6].nh == IPv6NH:
                            if type == 4:
                                if pkt.haslayer(IPv6ExtHdrSegmentRouting):
                                    if pkt[IPv6ExtHdrSegmentRouting].nh == IPv6ExtHdrRoutingNH:
                                        if pkt[IPv6ExtHdrSegmentRouting].type == type:
                                            if pkt[IPv6ExtHdrSegmentRouting].segleft == 0: 
                                                if pkt[IPv6ExtHdrSegmentRouting].addresses == [self.dstIP]:
                                                    toInsert = True
                            else:
                                if pkt.haslayer(IPv6ExtHdrRouting):
                                    if pkt[IPv6ExtHdrRouting].nh == IPv6ExtHdrRoutingNH:
                                        segleft = 0
                                        if type == 2:
                                            segleft = 1
                                        if pkt[IPv6ExtHdrRouting].segleft == segleft:
                                            if pkt[IPv6ExtHdrRouting].type == type:
                                                if type == 0 or type == 1 or type == 2 or type == 3:
                                                    if pkt[IPv6ExtHdrRouting].addresses == [self.dstIP]:
                                                        toInsert = True
                                                elif type == 5:
                                                    if pkt[IPv6ExtHdrRouting].reserved == 0x00080000:
                                                        toInsert = True
                                                elif type == 6:
                                                    if pkt[IPv6ExtHdrRouting].reserved == 0x00000008:
                                                        toInsert = True
                        if toInsert:
                            validedProbes.append(pkt)
                
                # Retrieve responses
                for pkt in respPkts:
                    if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                        if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                            rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                            toInsert = False
                            if pkt[IPerror6].nh == IPv6NH:
                                if type == 4:
                                    if pkt.haslayer(IPv6ExtHdrSegmentRouting):
                                        if pkt[IPv6ExtHdrSegmentRouting].nh == IPv6ExtHdrRoutingNH:
                                            if pkt[IPv6ExtHdrSegmentRouting].type == type:
                                                if pkt[IPv6ExtHdrSegmentRouting].segleft == 0: 
                                                    if pkt[IPv6ExtHdrSegmentRouting].addresses == [self.dstIP]:
                                                        toInsert = True
                                else:
                                    if pkt.haslayer(IPv6ExtHdrRouting):
                                        if pkt[IPv6ExtHdrRouting].nh == IPv6ExtHdrRoutingNH:
                                            segleft = 0
                                            if type == 2:
                                                segleft = 1
                                            if pkt[IPv6ExtHdrRouting].segleft == segleft:
                                                if pkt[IPv6ExtHdrRouting].type == type:
                                                    if type == 0 or type == 1 or type == 2 or type == 3:
                                                        if pkt[IPv6ExtHdrRouting].addresses == [self.dstIP]:
                                                            toInsert = True
                                                    elif type == 5:
                                                        if pkt[IPv6ExtHdrRouting].reserved == 0x00080000:
                                                            toInsert = True
                                                    elif type == 6:
                                                        if pkt[IPv6ExtHdrRouting].reserved == 0x00000008:
                                                            toInsert = True
                            if toInsert:
                                validedICMPResps[rcvID] = pkt
                            else:
                                filteredICMPResps[rcvID] = pkt

                        elif pkt.haslayer(TCP):
                            if pkt[TCP].flags == 'SA':
                                rcvID = pkt[TCP].ack - 1
                                validedSAResps[rcvID] = pkt
                    
                # Match probe with corresponding response
                hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
                
                # Record results
                results.append({
                    'expName': "{}Type{}{}".format(expConfig['name'], type, proto),
                    'nbProbesSnd': maxHlimSeen,
                    'hops': hops, 
                })
                        
        return results
    
    """
    Analyse pkts related to fragmentation header experiments
    """
    def FragExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys() or not 'types' in expConfig.keys() or not 'm' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for type in expConfig['types']:
            for proto in expConfig['proto']:
                for m in expConfig['m']:
                    validedProbes = []
                    validedICMPResps = {}
                    filteredICMPResps = {}
                    validedSAResps = {}

                    # Configure variables
                    IPv6NH = 44
                    if proto == 'UDP':
                        traceProbes = self.traceProbesUDP
                        respPkts = self.respUDP
                        IPv6ExtHdrFragmentNH = 17
                    elif proto == 'TCP':
                        traceProbes = self.traceProbesTCP
                        respPkts = self.respTCP
                        IPv6ExtHdrFragmentNH = 6
                    else:
                        print('[ERROR] Unknow protocol')
                        continue
                    
                    # Compute full expName
                    expName = "frag" + str(type) + "m" + str(m)

                    # Retrieve traceroute start and end time
                    traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
                    if traceInfos is None:
                        return None
                    
                    startTime = float(traceInfos['START_TIME'])
                    endTime = float(traceInfos['END_TIME'])

                    # Retrieve probes
                    for pkt in traceProbes:
                        if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                            if pkt[IPv6].nh == IPv6NH and pkt[IPv6].plen == int(type) - 54:
                                if pkt[IPv6ExtHdrFragment].nh == IPv6ExtHdrFragmentNH and pkt[IPv6ExtHdrFragment].m == m:
                                    validedProbes.append(pkt)
                    
                    # Retrieve responses
                    for pkt in respPkts:
                        if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                            if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                                rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                toInsert = False
                                if pkt[IPerror6].nh == IPv6NH and pkt[IPerror6].plen == int(type) - 54:
                                    if pkt[IPv6ExtHdrFragment].nh == IPv6ExtHdrFragmentNH and pkt[IPv6ExtHdrFragment].m == m:
                                        toInsert = True
                                if toInsert:
                                    validedICMPResps[rcvID] = pkt
                                else:
                                    filteredICMPResps[rcvID] = pkt

                            elif pkt.haslayer(TCP):
                                if pkt[TCP].flags == 'SA':
                                    rcvID = pkt[TCP].ack - 1
                                    validedSAResps[rcvID] = pkt
                        
                    # Match probe with corresponding response
                    hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
                    
                    # Record results
                    results.append({
                        'expName': "{}{}m{}{}".format(expConfig['name'], type, m, proto),
                        'nbProbesSnd': maxHlimSeen,
                        'hops': hops, 
                    })
        
        return results
    
    """
    Analyse pkts related to authentication header experiments
    """
    def AHExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedProbes = []
            validedICMPResps = {}
            filteredICMPResps = {}
            validedSAResps = {}

            # Configure variables
            traceProbes = self.otherPkts
            respPkts = self.otherPkts
            IPv6NH = 51
            if proto == 'UDP':
                AHNH = 17
            elif proto == 'TCP':
                AHNH = 6
            else:
                print('[ERROR] Unknow protocol')
                continue

            # Compute full expName
            expName = "ah"

            # Retrieve traceroute start and end time
            traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
            if traceInfos is None:
                return None
            
            startTime = float(traceInfos['START_TIME'])
            endTime = float(traceInfos['END_TIME'])

            # Retrieve probes
            for pkt in traceProbes:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(IPv6) and pkt.haslayer(Raw):
                        if RAW_DATA in pkt[Raw].load:
                            if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and 1 <= pkt[IPv6].hlim <= 100:
                                if pkt[IPv6].nh == IPv6NH and pkt.haslayer(AH):
                                    if pkt[AH].nh == AHNH:
                                        if struct.pack(">H", self.sport) == pkt[Raw].load[0:2] and struct.pack(">H", self.dport) == pkt[Raw].load[2:4]:
                                            validedProbes.append(pkt)
            
            # Retrieve responses
            for pkt in respPkts:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        if pkt.haslayer(IPerror6) and pkt.haslayer(Raw):
                            if RAW_DATA in pkt[Raw].load:
                                if pkt[IPerror6].src == self.srcIP and pkt[IPerror6].dst == self.dstIP:
                                    rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                    toInsert = False
                                    if pkt[IPerror6].nh == IPv6NH and pkt.haslayer(AH):
                                        if pkt[AH].nh == AHNH:
                                            if struct.pack(">H", self.sport) == pkt[Raw].load[0:2] and struct.pack(">H", self.dport) == pkt[Raw].load[2:4]:
                                                toInsert = True
                                    if toInsert:
                                        validedICMPResps[rcvID] = pkt
                                    else:
                                        filteredICMPResps[rcvID] = pkt
                    elif pkt.haslayer(TCP):
                        if pkt[TCP].flags == 'SA':
                            rcvID = pkt[TCP].ack - 1
                            validedSAResps[rcvID] = pkt
                
            # Match probe with corresponding response
            hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
            
            # Record results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'nbProbesSnd': maxHlimSeen,
                'hops': hops, 
            })
            
        return results
    
    """
    Analyse pkts related to ESP header experiments
    """
    def ESPExtHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedProbes = []
            validedICMPResps = {}
            filteredICMPResps = {}
            validedSAResps = {}

            # Configure variables
            
            IPv6NH = 50
            if proto == 'UDP' or proto == 'TCP':
                traceProbes = self.otherPkts
                respPkts = self.otherPkts
            else:
                print('[ERROR] Unknow protocol')
                continue

            # Compute full expName
            expName = "esp"

            # Retrieve traceroute start and end time
            traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto.lower(), TRACEINFOSFILENAME), expName)
            if traceInfos is None:
                return None
            
            startTime = float(traceInfos['START_TIME'])
            endTime = float(traceInfos['END_TIME'])

            # Retrieve probes
            for pkt in traceProbes:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(IPv6):
                        if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and 1 <= pkt[IPv6].hlim <= 100:
                            if pkt[IPv6].nh == IPv6NH and pkt.haslayer(ESP):
                                validedProbes.append(pkt)
            
            # Retrieve responses
            for pkt in respPkts:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        if pkt.haslayer(IPerror6):
                            if pkt[IPerror6].src == self.srcIP and pkt[IPerror6].dst == self.dstIP:
                                if pkt[IPerror6].nh == IPv6NH and pkt.haslayer(ESP):
                                    rcvID = int(pkt[ESP].seq)
                                    validedICMPResps[rcvID] = pkt
                                    
                    elif pkt.haslayer(TCP):
                        if pkt[TCP].flags == 'SA':
                            rcvID = pkt[TCP].ack - 1
                            validedSAResps[rcvID] = pkt
                
            # Match probe with corresponding response
            hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
            
            # Record results
            results.append({
                'expName': "{}{}".format(expConfig['name'], proto),
                'nbProbesSnd': maxHlimSeen,
                'hops': hops, 
            })
            
        return results

    """
    Analyse pkts related to no next header experiments
    """
    def NoNextHdr_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedProbes = []
            validedICMPResps = {}
            filteredICMPResps = {}
            validedSAResps = {}

            # Configure variables
            traceProbes = self.otherPkts
            respPkts = self.otherPkts
            if proto == 'noL4':
                IPv6NH = 59
            else:
                print('[ERROR] Unknow protocol')
                continue

            # Compute full expName
            expName = "noNextHdr"

            # Retrieve traceroute start and end time
            traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto, TRACEINFOSFILENAME), expName)
            if traceInfos is None:
                return None
            
            startTime = float(traceInfos['START_TIME'])
            endTime = float(traceInfos['END_TIME'])

            # Retrieve probes
            for pkt in traceProbes:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(IPv6) and pkt.haslayer(Raw):
                        if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and 1 <= pkt[IPv6].hlim <= 100:
                            if pkt [IPv6].nh == IPv6NH:
                                if RAW_DATA in pkt[Raw].load:
                                    validedProbes.append(pkt)
            
            # Retrieve responses
            for pkt in respPkts:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        if pkt.haslayer(IPerror6) and pkt.haslayer(Raw):
                            if RAW_DATA in pkt[Raw].load:
                                if pkt[IPerror6].src == self.srcIP and pkt[IPerror6].dst == self.dstIP:
                                    rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                    toInsert = False
                                    if pkt[IPerror6].nh == IPv6NH:
                                        toInsert = True
                                    if toInsert:
                                        validedICMPResps[rcvID] = pkt
                                    else:
                                        filteredICMPResps[rcvID] = pkt
                    elif pkt.haslayer(TCP):
                        if pkt[TCP].flags == 'SA':
                            rcvID = pkt[TCP].ack - 1
                            validedSAResps[rcvID] = pkt
                
            # Match probe with corresponding response
            hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
            
            # Record results
            results.append({
                'expName': "{}".format(expConfig['name']),
                'nbProbesSnd': maxHlimSeen,
                'hops': hops, 
            })
            
        return results
    
    """
    Analyse pkts related to ethernet header experiments
    """
    def Ethernet_analyse(self, expConfig):
        if not 'proto' in expConfig.keys():
            return
        
        results = []
        # Iterate through the different experiments
        for proto in expConfig['proto']:
            validedProbes = []
            validedICMPResps = {}
            filteredICMPResps = {}
            validedSAResps = {}

            # Configure variables
            traceProbes = self.otherPkts
            respPkts = self.otherPkts
            if proto == 'noL4':
                IPv6NH = 143
            else:
                print('[ERROR] Unknow protocol')
                continue

            # Compute full expName
            expName = "ethernet"

            # Retrieve traceroute start and end time
            traceInfos = retrieve_traceroute_infos("{}{}/{}".format(self.traceInfosPath, proto, TRACEINFOSFILENAME), expName)
            if traceInfos is None:
                return None
            
            startTime = float(traceInfos['START_TIME'])
            endTime = float(traceInfos['END_TIME'])

            # Retrieve probes
            for pkt in traceProbes:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(IPv6) and pkt.haslayer(Raw):
                        if pkt[IPv6].src == self.srcIP and pkt[IPv6].dst == self.dstIP and 1 <= pkt[IPv6].hlim <= 100:
                            if pkt[IPv6].nh == IPv6NH:
                                if RAW_DATA in pkt[Raw].load:
                                    validedProbes.append(pkt)
            
            # Retrieve responses
            for pkt in respPkts:
                if (startTime - DELTATIME) <= pkt.time <= (endTime + DELTATIME):
                    if pkt.haslayer(ICMPv6TimeExceeded) or pkt.haslayer(ICMPv6ParamProblem) or pkt.haslayer(ICMPv6DestUnreach):
                        if pkt.haslayer(IPerror6) and pkt.haslayer(Raw):
                            if RAW_DATA in pkt[Raw].load:
                                if pkt[IPerror6].src == self.srcIP and pkt[IPerror6].dst == self.dstIP:
                                    rcvID = int(pkt[Raw].load.partition(RAW_DATA)[2].partition(SPLITBYTE)[0])
                                    toInsert = False
                                    if pkt[IPerror6].nh == IPv6NH:
                                        toInsert = True
                                    if toInsert:
                                        validedICMPResps[rcvID] = pkt
                                    else:
                                        filteredICMPResps[rcvID] = pkt
                    elif pkt.haslayer(TCP):
                        if pkt[TCP].flags == 'SA':
                            rcvID = pkt[TCP].ack - 1
                            validedSAResps[rcvID] = pkt
                
            # Match probe with corresponding response
            hops, maxHlimSeen = self.match_probes_responses(validedProbes, validedICMPResps, filteredICMPResps, validedSAResps)
            
            # Record results
            results.append({
                'expName': "{}".format(expConfig['name']),
                'nbProbesSnd': maxHlimSeen,
                'hops': hops, 
            })
            
        return results
