"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import utils
import os
import config as config
from scapy.all import *
from modules.experiments_analyse import ExperimentsAnalyse
from modules.traceroutes_analyse import TracerouteAnalyse

HOMEDSRCRSTSDIR = "tests_results"
RCVDATAFILE = "received_traffic.pcap"
PATHCPTDATADEST = "captures/{}_capture.pcap"
EXPSINPUTFILE = "experiments_to_analyse.ndjson"
ASINFOCACHEFILE = "caches/ASInfo_cache.ndjson"
ASTYPECACHEFILE = "caches/ASType_cache.ndjson"
NOTMATCHIPSAVEFILE = "not_match_IP.txt"
SEENIPSSAVEFILE = "seen_IPs.txt"
HOPSNBMATRIXSAVEDIR = "hopsMatrix"
EXPRSLTSAVEFILE = "SD_summary.txt"
TESTSUMSAVEFILE = "test_summary.txt"
SAVEDIRNAME = "analysed_results"
HOSTFILE = "hosts_infos.txt"
SDDIRPATH = "SD"
HOPSCNTTOFILE = "hopCntMatrixTo_{}.txt"
COMPRESSASINFOSFILE = "analysed_AS.txt"
UDPCOMPUTEASTRACEFROM = ['StartTraceUDP', 'NoHeaderUDP', 'EndTraceUDP']
TCPCOMPUTEASTRACEFROM = ['StartTraceTCP', 'NoHeaderTCP', 'EndTraceTCP']
ASTYPECACHE = {}
ASINFOSCACHE = {
    'cache': {},
    'tree': {},
}
IPSINFOS = {
    'seenIPs': {},
    'notMatchIPs': {},
}

"""
Determine AS Trace
"""
def determine_ASTrace(srcDstRslts, testsRslts, SDName, dstIP):
    for proto in config.PROTOCOLS:
        hlimDest = None
        computeSet = UDPCOMPUTEASTRACEFROM if proto == 'udp' else TCPCOMPUTEASTRACEFROM

        # Compute and combine ASTrace from the specified experiments
        for expName in computeSet:
            SDASTrace = None
            hops = srcDstRslts['tracesRsltsByExp'][expName]['hops']
            hopsCnt = srcDstRslts['hopsCntByExp'][expName]
            
            SDASTrace = utils.compute_ASTrace(dstIP, hops, hopsCnt, ASINFOSCACHE, IPSINFOS)

            # Insert seen AS
            for asn, infos in SDASTrace['seenAS'].items():
                if asn not in testsRslts['seenAS']:
                    testsRslts['seenAS'][asn] = infos

            # Insert ASTrace
            if SDName not in testsRslts['ASTraceByProto'][proto]:
                # Find destination hlim
                for hlim, hopInfos in SDASTrace['ASTrace'].items():
                    if hopInfos[1] is True:
                        hlimDest = hlim
                
                testsRslts['ASTraceByProto'][proto][SDName] = SDASTrace
            else:
                # Combine ASTrace
                for hlim, hopInfos in SDASTrace['ASTrace'].items():
                    if hlim not in testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']:
                        # Dest not reached yet
                        if hlimDest is None:
                            testsRslts['ASTraceByProto'][proto][SDName]['ASTrace'][hlim] = hopInfos
                            if hopInfos[1] is True:
                                hlimDest = hlim
                        else:
                            if hlim < hlimDest and hopInfos[1] is False:
                                testsRslts['ASTraceByProto'][proto][SDName]['ASTrace'][hlim] = hopInfos
            
        # Compute new tracelen
        SDASTrace = testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']
        testsRslts['ASTraceByProto'][proto][SDName]['traceLen'] = max(SDASTrace.keys())

        # Compute new ASPath
        ASPath = []
        lastASNSeen = None
        for hlim, hopInfos in SDASTrace.items():
            if type(hopInfos[0]) is dict:
                # Insert in ASPath
                if hopInfos[0]['asn'] != lastASNSeen:
                    ASPath.append(hopInfos[0]['asn'])
                lastASNSeen = hopInfos[0]['asn']

        testsRslts['ASTraceByProto'][proto][SDName]['ASPath'] = ASPath
    
    return

"""
Perform requested analyses for a given source-destination
"""
def perform_analyzes_src_dest(srcDirPath, srcHostName, srcHostInfo, dstHostName, dstHostInfo, savePath, exps, rcvPkts):
    print("  From {} to {} --> ANALYSING".format(srcHostName.upper(), dstHostName.upper()), end='\r', flush=True)

    # Load sended pkts
    testsDataDirPath = "{}/{}/".format(srcDirPath, HOMEDSRCRSTSDIR)
    sndPkts = rdpcap(testsDataDirPath + PATHCPTDATADEST.format(dstHostName.upper()))

    # Filter pkts to match source destination
    filterSndPkts, filterRcvPkts = utils.filter_pkts(srcHostInfo['ip'], sndPkts, rcvPkts)
    
    # Perform analyses
    expsAnalyseRslts = ExperimentsAnalyse(
        srcHostInfo['ip'], 
        dstHostInfo['ip'], 
        config.DEFAULT_SPORT, 
        dstHostInfo['openPort'], 
        filterSndPkts, 
        filterRcvPkts, 
        exps
        ).start_experiments_analyse()

    tracesAnalyseRslts = TracerouteAnalyse(
        srcHostInfo['ip'], 
        dstHostInfo['ip'], 
        config.DEFAULT_SPORT, 
        dstHostInfo['openPort'], 
        filterSndPkts, 
        exps, 
        testsDataDirPath + dstHostName.upper() + '/'
        ).start_traceroutes_analyse()

    # Summarize results
    summary = {
        'SD': srcHostName.upper() + dstHostName.upper(), 
        'expRsltsByExp': {}, 
        'tracesRsltsByExp': {},
        'hopsCntByExp': {}
    }

    # Record experiments results
    for expName, data in expsAnalyseRslts.items():
        # Compare experiment median delay with no ext header median delay
        noHeaderExpName = 'NoHeaderTCP' if expName[-3:] == 'TCP' else 'NoHeaderUDP'
        noHeaderDelay = expsAnalyseRslts[noHeaderExpName]['medianDelay']
        expDelay = data['medianDelay']
        if expDelay is None or noHeaderDelay is None:
            deltaDelay = None
        else:
            deltaDelay = round(expDelay - noHeaderDelay, 2)
        
        # Insert data
        summary['expRsltsByExp'][expName] = {'expData': data, 'deltaDelay': deltaDelay}

    # Record traceroutes results
    for expName, data in tracesAnalyseRslts.items():
        summary['tracesRsltsByExp'][expName] = data
        summary['hopsCntByExp'][expName] = max(data['hops'].keys()) if len(data['hops'].keys()) > 0 else 0

    # Save results
    utils.save_results_expsAnalyse(summary['expRsltsByExp'], savePath, EXPRSLTSAVEFILE)
    utils.save_results_traceAnalyse(summary['tracesRsltsByExp'], savePath, srcHostInfo['ip'], dstHostInfo['ip'])

    print("  From {} to {} --> COMPLETE    ".format(srcHostName.upper(), dstHostName.upper()))

    return summary

"""
Perform analyses on every source-destination
"""
def perform_analyses(rawDataPath, hosts, selectedHostsNames, savePath, exps):
    testsRslts = {
        'analysedAS': {},
        'seenAS': {},
        'ASTraceByProto': {'udp': {}, 'tcp': {}},
        'nbSDTested': 0,
        'nbSDUDPReached': 0,
        'nbSDTCPReached': 0,
        'dataByExp': {}
    }

    # Iterate through destination
    for dstHostName in selectedHostsNames:
        hopsCntMatrix = []
        hopsCntMatrixSavePath = '{}/{}/'.format(SAVEDIRNAME, HOPSNBMATRIXSAVEDIR)
        hopsCntMatrixSaveFile = HOPSCNTTOFILE.format(dstHostName)
        dstDirPath = '{}/{}'.format(rawDataPath, dstHostName)
        dstRawDataFile = '{}/{}'.format(dstDirPath, RCVDATAFILE)

        # Compute the list of source 
        srcHostsNames = selectedHostsNames[:]
        srcHostsNames.remove(dstHostName)

        # Load destination received file
        rcvPkts = rdpcap(dstRawDataFile)

        # Iterate through source
        for srcHostName in srcHostsNames:
            srcDirPath = '{}/{}'.format(rawDataPath, srcHostName)
            SDDir = srcHostName.upper() + dstHostName.upper()
            saveSDPath = '{}{}/'.format(savePath, SDDir)

            # Perform the analyses for source-destination
            srcDstRslts = perform_analyzes_src_dest(srcDirPath, srcHostName, hosts[srcHostName], dstHostName, hosts[dstHostName], saveSDPath, exps, rcvPkts)
            
            # Determine ASTrace
            SDName = srcDstRslts['SD']
            determine_ASTrace(srcDstRslts, testsRslts, SDName, hosts[dstHostName]['ip'])

            # Check if ASTrace reached destination for each protocol
            ASTraceReachDstByProto = {}
            for proto in config.PROTOCOLS:
                ASTrace = testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']
                isDstReached = False

                for hlim, hopInfo in ASTrace.items():
                    if hopInfo[1] is True:
                        isDstReached = True
                        break

                if isDstReached:
                    ASTraceReachDstByProto[proto] = True
                    if proto == 'udp':
                        testsRslts['nbSDUDPReached'] += 1
                    else:
                        testsRslts['nbSDTCPReached'] += 1
                else:
                    ASTraceReachDstByProto[proto] = False
            
            # Analyse every experiments
            expsList = srcDstRslts['expRsltsByExp'].keys()
            for expName in expsList:
                # Create a experiments summary dictionnary
                if expName not in testsRslts['dataByExp']:
                    testsRslts['dataByExp'][expName] = {
                        'nbTests': 0,
                        'nbPktsSend': 0,
                        'nbPktsRcv': 0,
                        'nbPktsFiltered': 0,
                        'expsSmryBySD': {},
                        'tracesFailure': [],
                    }
                
                sndPktsCnt = srcDstRslts['expRsltsByExp'][expName]['expData']['sndPktsCnt']
                rcvPktsCnt = srcDstRslts['expRsltsByExp'][expName]['expData']['rcvPktsCnt']
                filteredPktsCnt = srcDstRslts['expRsltsByExp'][expName]['expData']['filteredPktsCnt']
                state = srcDstRslts['expRsltsByExp'][expName]['expData']['state']
                expDelay = srcDstRslts['expRsltsByExp'][expName]['expData']['medianDelay']
                deltaDelay = srcDstRslts['expRsltsByExp'][expName]['deltaDelay']
                expSmryTuple = (sndPktsCnt, rcvPktsCnt, filteredPktsCnt, state, expDelay, deltaDelay)

                # Record into summary dictionay
                testsRslts['dataByExp'][expName]['nbTests'] += 1
                testsRslts['dataByExp'][expName]['nbPktsSend'] += expSmryTuple[0]
                testsRslts['dataByExp'][expName]['nbPktsRcv'] += expSmryTuple[1]
                testsRslts['dataByExp'][expName]['nbPktsFiltered'] += expSmryTuple[2]
                testsRslts['dataByExp'][expName]['expsSmryBySD'][SDName] = expSmryTuple

                problemType = None
                proto = 'tcp' if expName[-3:] == 'TCP' else 'udp'
                hops = srcDstRslts['tracesRsltsByExp'][expName]['hops']
                lastHopIP = hops[srcDstRslts['hopsCntByExp'][expName]]['ip'] if srcDstRslts['hopsCntByExp'][expName] is not 0 else None
                expRcvPktsCnt = srcDstRslts['expRsltsByExp'][expName]['expData']['rcvPktsCnt']
                expFiltPktsCnt = srcDstRslts['expRsltsByExp'][expName]['expData']['filteredPktsCnt']
                if 'UDP' in expName:
                    SDASTrace = testsRslts['ASTraceByProto']['udp'][SDName]['ASTrace']
                else:
                    SDASTrace = testsRslts['ASTraceByProto']['tcp'][SDName]['ASTrace']

                # Check if exp is filtered
                if expFiltPktsCnt != 0:
                    problemType = 'filtering'
                    ASInfos = utils.find_firstFiltering_AS(hops, SDASTrace, ASINFOSCACHE, IPSINFOS)

                # Check if no packets are received by destination
                if expRcvPktsCnt == 0 and expFiltPktsCnt == 0:
                    # Check if ASTrace is complete and that exp last hop is not the destination
                    if ASTraceReachDstByProto[proto] == True and lastHopIP != hosts[dstHostName]['ip']:
                        # Try to find the failing AS
                        problemType = 'blocking'
                        ASInfos = utils.determine_failingAS(hops, SDASTrace, ASINFOSCACHE, IPSINFOS)

                if problemType is not None:
                    # Record AS if not already seen
                    if ASInfos is not None and type(ASInfos) is dict:
                        if ASInfos['asn'] not in testsRslts['seenAS']:
                            testsRslts['seenAS'][ASInfos['asn']] = ASInfos
                        if ASInfos['asn'] not in testsRslts['analysedAS']:
                            testsRslts['analysedAS'][ASInfos['asn']] = ASInfos['asn_description']

                    # Record failingASInfo
                    testsRslts['dataByExp'][expName]['tracesFailure'].append({
                            'SD': srcDstRslts['SD'],
                            'failingASInfo': ASInfos if type(ASInfos) is dict else None,
                            'problemType': problemType
                    })
                
                # Record analysed AS
                ASPath = testsRslts['ASTraceByProto'][proto][SDName]['ASPath']
                for asn in ASPath:
                    if asn in testsRslts['seenAS']:
                        desc = testsRslts['seenAS'][asn]['asn_description']
                        testsRslts['analysedAS'][asn] = desc

            # Insert into hops matrix
            hopsCntMatrix.append({
                'SD': SDName,
                'hopsCntByExp': srcDstRslts['hopsCntByExp']
            })

            # Record statistics
            testsRslts['nbSDTested'] += 1

        # Save hops matrix
        utils.save_HopsCntMatrix(hopsCntMatrix, testsRslts['dataByExp'].keys(), hopsCntMatrixSavePath, hopsCntMatrixSaveFile)
                   
    # Save data
    utils.save_summary_test(testsRslts, savePath, TESTSUMSAVEFILE)

    return testsRslts

"""
Start analyses
"""
def main():
    try:
        # Load all the experiments information to analyse
        exps = utils.load_experiments(EXPSINPUTFILE)
        # Load hosts
        hosts = utils.load_hosts(HOSTFILE)
        # Load AS caches
        utils.load_ASCaches(ASINFOCACHEFILE, ASTYPECACHEFILE, ASINFOSCACHE, ASTYPECACHE)
    except:
        print("[ERROR] Unable to load required data")
        exit(1)

    blockingAS = {}
    filteringAS = {}
    delayInfosByExp = {}
    analyseRslts = {
        'nbSDTested': 0,
        'nbSDUDPReached': 0,
        'nbSDTCPReached': 0,
        'dataByExp': {},
    }

    # Compute savePath
    savePath = "{}/{}/".format(SAVEDIRNAME, SDDIRPATH)

    # Find used machines
    dirNames = os.listdir(config.RAWDATA_PATH)
    selectedVMs = []
    for name in dirNames:
        if name in hosts:
            selectedVMs.append(name)
    
    print('---- Performing analyses ----')
    # Perform analyses for a given round
    results = perform_analyses(config.RAWDATA_PATH, hosts, selectedVMs, savePath, exps)
    print(''.ljust(27, '-'))

    # Record general data
    analyseRslts['nbSDTested'] = results['nbSDTested']
    analyseRslts['nbSDUDPReached'] = results['nbSDUDPReached']
    analyseRslts['nbSDTCPReached'] = results['nbSDTCPReached']

    # Record data for each experiment
    for expName, expData in results['dataByExp'].items():
        if expName not in analyseRslts['dataByExp']:
            analyseRslts['dataByExp'][expName] = {
                'sndPktsTotal': 0,
                'rcvPktsTotal': 0,
                'filteredPktsTotal': 0
            }
        
        # Record send, receive, and filtered pkts count
        analyseRslts['dataByExp'][expName]['sndPktsTotal'] += expData['nbPktsSend']
        analyseRslts['dataByExp'][expName]['rcvPktsTotal'] += expData['nbPktsRcv']
        analyseRslts['dataByExp'][expName]['filteredPktsTotal'] += expData['nbPktsFiltered']
        
        # Collect experiment recorded for all SD
        delayList = []
        for SDName, smry in expData['expsSmryBySD'].items():
            SDExpDelay = smry[4]
            if SDExpDelay is not None and 0 < SDExpDelay <= 1000:
                delayList.append(smry[4])

        # Compute statistics on experiment delay
        delayInfosByExp[expName] = utils.compute_delayInfos(delayList)

        # Compute delta delay compared to NoHeader exp (comparing median)
        noHeaderExpName = 'NoHeaderTCP' if expName[-3:] == 'TCP' else 'NoHeaderUDP'
        if delayInfosByExp[noHeaderExpName]['median'] is not None and delayInfosByExp[expName]['median'] is not None:
            noHeaderDelay = delayInfosByExp[noHeaderExpName]['median']
            expDelay = delayInfosByExp[expName]['median']
            delayInfosByExp[expName]['delta'] = round(expDelay - noHeaderDelay,2)
        
        # Collect failed SD
        if len(expData['tracesFailure']) != 0:
            for SDFail in expData['tracesFailure']:
                failingSD = SDFail['SD']

                # Record filtering AS
                if SDFail['problemType'] == 'filtering':
                    ASSet = filteringAS
                else:
                    # Record dropping AS
                    ASSet = blockingAS

                if SDFail['failingASInfo'] is not None and type(SDFail['failingASInfo']) is dict:
                    failingASN = SDFail['failingASInfo']['asn']
                    if failingASN not in ASSet:
                        ASSet[SDFail['failingASInfo']['asn']] = {
                            'ASinfo': SDFail['failingASInfo'],
                            'exps': {expName: [failingSD]},
                        }
                    else:
                        if expName not in ASSet[failingASN]['exps']:
                            ASSet[failingASN]['exps'][expName] = [failingSD]
                        else:
                            ASSet[failingASN]['exps'][expName].append(failingSD)
    
    # Save delay infos
    utils.save_delayInfos(delayInfosByExp, SAVEDIRNAME, 'delay_summary.txt')
    # Save results
    utils.save_data_summary(analyseRslts, SAVEDIRNAME, 'data_summary.txt')
    # Save analysed AS and blocking AS
    utils.save_ASInfos(results['seenAS'], results['analysedAS'], blockingAS, filteringAS, SAVEDIRNAME, 'seen_AS.txt', 'blocking_AS.txt', 'filtering_AS.txt', COMPRESSASINFOSFILE, ASTYPECACHE)
    # Save ASPath udp
    utils.save_ASPath(results['ASTraceByProto']['udp'], SAVEDIRNAME, 'AS_path_udp.txt')
    # Save ASPath tcp
    utils.save_ASPath(results['ASTraceByProto']['tcp'], SAVEDIRNAME, 'AS_path_tcp.txt')
    # Save IPs information
    utils.save_IPsInfos(IPSINFOS, SAVEDIRNAME, SEENIPSSAVEFILE, NOTMATCHIPSAVEFILE)
    # Saving ASInfoCache
    utils.save_ASCache(ASINFOSCACHE, ASINFOCACHEFILE)

    return 0
    
if __name__ == "__main__":
    main()
