"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ipaddress
import utils
import os
import config as config
from scapy.all import *
from modules.traceroutes_analyse import TracerouteAnalyse

HOMEDSRCRSTSDIR = "tests_results"
RCVDATAFILE = "received_traffic.pcap"
PATHCPTDATADEST = "captures/{}_capture.pcap"
EXPSINPUTFILE = "experiments_to_analyse.ndjson"
ASINFOCACHEFILE = "caches/ASInfo_cache.ndjson"
ASTYPECACHEFILE = "caches/ASType_cache.ndjson"
NOTMATCHIPSAVEFILE = "not_match_IP.txt"
SEENIPSSAVEFILE = 'seen_IPs.txt'
HOPSNBMATRIXSAVEDIR = "hopsMatrix"
TESTSUMSAVEFILE = "test_summary.txt"
SAVEDIRNAME = "analysed_results"
HOSTFILE = "hosts_infos.txt"
SDDIRPATH = "SD"
COMPRESSASINFOSFILE = "analysed_AS.txt"
HOPCNTTOFILE = "hopCntMatrixTo_{}.txt"
CONFIGFILE = 'IPv6_dstAddr.txt'
UDPCOMPUTEASTRACEFROM = ['StartTraceUDP', 'NoHeaderUDP', 'EndTraceUDP']
TCPCOMPUTEASTRACEFROM = ['StartTraceTCP', 'NoHeaderTCP', 'EndTraceTCP']
ASTYPECACHE = {}
ASINFOSCACHE = {
    'cache': {},
    'tree': {},
}
IPSINFOS = {
    'seenIPs': {},
    'notMatchIPs': {},
}

"""
Determine AS Trace
"""
def determine_ASTrace(srcDstRslts, testsRslts, SDName, dstASN, dstPrefix):
    for proto in config.PROTOCOLS:
        hlimDest = None
        computeSet = UDPCOMPUTEASTRACEFROM if proto == 'udp' else TCPCOMPUTEASTRACEFROM

        # Compute and combine ASTrace from the specified experiments
        for expName in computeSet:
            SDASTrace = None
            hops = srcDstRslts['tracesRsltsByExp'][expName]['hops']
            hopsCnt = srcDstRslts['hopsCntByExp'][expName]

            SDASTrace = utils.compute_ASTrace(dstASN, dstPrefix, hops, hopsCnt, ASINFOSCACHE, IPSINFOS)

            # Combine seen AS
            for asn, infos in SDASTrace['seenAS'].items():
                if asn not in testsRslts['seenAS']:
                    testsRslts['seenAS'][asn] = infos
            
            # Combine ASTrace
            if SDName not in testsRslts['ASTraceByProto'][proto]:
                # Find destination hlim
                for hlim, hopInfos in SDASTrace['ASTrace'].items():
                    if hopInfos[1] is True:
                        hlimDest = hlim
                
                testsRslts['ASTraceByProto'][proto][SDName] = SDASTrace
            else:
                # Combine ASTrace
                for hlim, hopInfos in SDASTrace['ASTrace'].items():
                    if hlim not in testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']:
                        # Dest not reached yet
                        if hlimDest is None:
                            testsRslts['ASTraceByProto'][proto][SDName]['ASTrace'][hlim] = hopInfos
                            if hopInfos[1] is True:
                                hlimDest = hlim
                        else:
                            if hlim < hlimDest and hopInfos[1] is False:
                                testsRslts['ASTraceByProto'][proto][SDName]['ASTrace'][hlim] = hopInfos
        
        # Compute new tracelen
        SDASTrace = testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']
        testsRslts['ASTraceByProto'][proto][SDName]['traceLen'] = max(SDASTrace.keys()) if len(SDASTrace.keys()) != 0 else 0

        # Compute new ASPath
        ASPath = []
        lastASNSeen = None
        for hlim, hopInfos in SDASTrace.items():
            if type(hopInfos[0]) is dict:
                # Insert in ASPath
                if hopInfos[0]['asn'] != lastASNSeen:
                    ASPath.append(hopInfos[0]['asn'])
                lastASNSeen = hopInfos[0]['asn']

        testsRslts['ASTraceByProto'][proto][SDName]['ASPath'] = ASPath
    
    return

"""
Perform requested analyses for a given source-destination
"""
def perform_analyzes_src_dest(srcDirPath, srcHostName, srcHostInfo, dstHostName, dstHostInfo, originAS, dstPrefix, savePath, exps):
    print("  From {} to {} --> ANALYSING".format(srcHostName.upper(), dstHostName), end='\r', flush=True)

    # Load sended pkts
    testsDataDirPath = "{}/{}/".format(srcDirPath, HOMEDSRCRSTSDIR)
    sndPkts = rdpcap(testsDataDirPath + PATHCPTDATADEST.format(dstHostName))

    # Filter pkts to match source destination
    filterSndPkts, filterRcvPkts = utils.filter_pkts(srcHostInfo['ip'], sndPkts)

    # Analyse traceroutes
    tracesAnalyseRslts = TracerouteAnalyse(
        srcHostInfo['ip'],
        dstHostInfo['dstIP'],
        originAS,
        dstPrefix,
        config.DEFAULT_SPORT,
        dstHostInfo['dport'],
        filterSndPkts,
        exps,
        testsDataDirPath + dstHostName + '/', 
        ASINFOSCACHE,
        IPSINFOS
        ).start_traceroutes_analyse()

    # Summarize results
    summary = {
        'SD': srcHostName.upper() + dstHostName.upper(), 
        'tracesRsltsByExp': {},
        'hopsCntByExp': {}
    }
    
    # Record traceroutes results
    for expName, data in tracesAnalyseRslts.items():
        summary['tracesRsltsByExp'][expName] = data
        summary['hopsCntByExp'][expName] = max(data['hops'].keys()) if len(data['hops'].keys()) > 0 else 0

    # Save results
    utils.save_results_traceAnalyse(summary['tracesRsltsByExp'], savePath, srcHostInfo['ip'], dstHostInfo['dstIP'])

    print("  From {} to {} --> COMPLETE    ".format(srcHostName.upper(), dstHostName))

    return summary

"""
Perform analyses on every source-destination
"""
def perform_analyses(rawDataPath, hosts, usedHosts, destsInfos, savePath, exps):
    testsRslts = {
        'analysedAS': {},
        'seenAS': {},
        'ASTraceByProto': {'udp': {}, 'tcp': {}},
        'nbSDTested': 0,
        'nbSDUDPReached': 0,
        'nbSDTCPReached': 0,
        'dataByExp': {}
    }

    # Iterate through hosts directories
    for srcHostName in usedHosts:
        srcDirPath = '{}/{}/'.format(rawDataPath, srcHostName)
        probedDestList = sorted(os.listdir(srcDirPath+HOMEDSRCRSTSDIR))
        
        # Remove captures from list if present
        if 'captures' in probedDestList:
            probedDestList.remove('captures')
        
        # Load src host config file
        configFilePath = '{}/{}/{}'.format(config.CONFIGDIR_PATH, srcHostName, CONFIGFILE)
        srcProbedDests = utils.retrieve_src_destinations(configFilePath)

        # Iterate through probed destinations
        for dstHostName in probedDestList:
            # Check if destination host is in config
            if dstHostName not in srcProbedDests:
                continue

            hopsCntMatrix = []
            hopsCntMatrixSavePath = '{}/{}/'.format(SAVEDIRNAME, HOPSNBMATRIXSAVEDIR)
            hopsCntMatrixSaveFile = HOPCNTTOFILE.format(dstHostName)
            SDDir = srcHostName.upper() + dstHostName
            saveSDPath = '{}{}/'.format(savePath, SDDir)            
            originAS = destsInfos[dstHostName]['originAS']
            dstPrefix = destsInfos[dstHostName]['cidr']

            # Perform the analyses for source-destination
            srcDstRslts = perform_analyzes_src_dest(
                srcDirPath, srcHostName, hosts[srcHostName],
                dstHostName, srcProbedDests[dstHostName], originAS, dstPrefix, saveSDPath, exps
            )
            
            # Determine ASTrace
            SDName = srcDstRslts['SD']
            determine_ASTrace(srcDstRslts, testsRslts, SDName, originAS, dstPrefix)

            # Check if ASTrace reached destination for each protocol
            ASTraceReachDstByProto = {}
            for proto in config.PROTOCOLS:
                ASTrace = testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']
                isDstReached = False

                for hlim, hopInfo in ASTrace.items():
                    if hopInfo[1] is True:
                        isDstReached = True
                        break

                if isDstReached:
                    ASTraceReachDstByProto[proto] = True
                    if proto == 'udp':
                        testsRslts['nbSDUDPReached'] += 1
                    else:
                        testsRslts['nbSDTCPReached'] += 1
                else:
                    ASTraceReachDstByProto[proto] = False
            
            # Remove StartTrace and EndTrace from exps list
            expsList = srcDstRslts['tracesRsltsByExp'].keys()
            expsList = [exp for exp in expsList if not (exp.startswith('StartTrace') or exp.startswith('EndTrace'))]
            
            for expName in expsList:
                # Check if AS Trace reached destination
                proto = 'tcp' if expName[-3:] == 'TCP' else 'udp'
                if ASTraceReachDstByProto[proto] == False:
                    continue
                
                # Create a experiment summary dictionnary
                if expName not in testsRslts['dataByExp']:
                    testsRslts['dataByExp'][expName] = {
                        'nbSDTotal': 0,
                        'nbSDForwarding': 0,
                        'nbSDBlocking': 0,
                        'nbSDFiltering': 0,
                        'tracesFailure': [],
                    }

                problemType = None
                hops = srcDstRslts['tracesRsltsByExp'][expName]['hops']
                lastHopIP = hops[srcDstRslts['hopsCntByExp'][expName]]['ip'] if srcDstRslts['hopsCntByExp'][expName] is not 0 else None
                hopInfo = utils.retrieve_ASInfo(lastHopIP, ASINFOSCACHE, IPSINFOS)
                lastHopASN = hopInfo['asn'] if hopInfo is not None else None
                SDASTrace = testsRslts['ASTraceByProto'][proto][SDName]['ASTrace']
                
                # Check if exp filtered
                filteringASInfos = utils.find_firstFiltering_AS(hops, SDASTrace, ASINFOSCACHE, IPSINFOS)
                if filteringASInfos is not None:
                    problemType = 'filtering'
                    
                # Check if exp is blocked
                IPv6Object = ipaddress.ip_address(lastHopIP) if lastHopIP is not None else None
                networkObject = ipaddress.ip_network(dstPrefix) if lastHopIP is not None else []
                if lastHopASN != originAS and (IPv6Object not in networkObject):
                    blockingASInfos = utils.determine_failingAS(hops, SDASTrace, ASINFOSCACHE, IPSINFOS)
                    if blockingASInfos is not True:
                        problemType = 'blocking'
                    else:
                        blockingASInfos = None

                # Record failingASInfo
                if problemType is not None:
                    if problemType == 'filtering':
                        ASInfos = filteringASInfos
                        testsRslts['dataByExp'][expName]['nbSDFiltering'] += 1
                    else:
                        ASInfos = blockingASInfos
                        testsRslts['dataByExp'][expName]['nbSDBlocking'] += 1
                    
                    testsRslts['dataByExp'][expName]['tracesFailure'].append({
                        'SD': srcDstRslts['SD'],
                        'failingASInfo': ASInfos if type(ASInfos) is dict else None,
                        'problemType': problemType
                    })
                else:
                    testsRslts['dataByExp'][expName]['nbSDForwarding'] += 1
                    
                # Record statistics
                testsRslts['dataByExp'][expName]['nbSDTotal'] += 1

                # Record analysed AS
                ASPath = testsRslts['ASTraceByProto'][proto][SDName]['ASPath']
                for asn in ASPath:
                    if asn in testsRslts['seenAS']:
                        desc = testsRslts['seenAS'][asn]['asn_description']
                        testsRslts['analysedAS'][asn] = desc

            # Insert into hops matrix
            hopsCntMatrix.append({
                'SD': SDName,
                'hopsCntByExp': srcDstRslts['hopsCntByExp']
            })

            # Record statistics
            testsRslts['nbSDTested'] += 1

            # Save hops matrix
            utils.save_HopsCntMatrix(hopsCntMatrix, testsRslts['dataByExp'].keys(), hopsCntMatrixSavePath, hopsCntMatrixSaveFile)
    
    # Save data
    utils.save_summary_test(testsRslts, savePath, TESTSUMSAVEFILE)

    return testsRslts

"""
Start analyses
"""
def main():
    try:
        # Load all the experiments information to analyse
        exps = utils.load_experiments(EXPSINPUTFILE)
        # Load source hosts
        hosts = utils.load_hosts(HOSTFILE)
        # Load AS caches
        utils.load_ASCaches(ASINFOCACHEFILE, ASTYPECACHEFILE, ASINFOSCACHE, ASTYPECACHE)
        # Load all destinations
        destsInfos = utils.load_destsInfos(config.DESTSINFOSFILE)
    except:
        print("[ERROR] Unable to load required data")
        exit(1)

    blockingAS = {}
    filteringAS = {}

    # Compute savePath
    savePath = "{}/{}/".format(SAVEDIRNAME, SDDIRPATH)

    # Find used machines
    dirNames = os.listdir(config.RAWDATA_PATH)
    usedHosts = []
    for name in dirNames:
        if name in hosts:
            usedHosts.append(name)

    print('---- Performing analyses ----')
    # Perform analyse for a given round
    results = perform_analyses(config.RAWDATA_PATH, hosts, usedHosts, destsInfos, savePath, exps)
    print(''.ljust(27, '-'))
    
    # Aggregate data for an experiment
    for expName, expData in results['dataByExp'].items():
        # Collect failed SD
        if len(expData['tracesFailure']) != 0:
            for SDFail in expData['tracesFailure']:
                failingSD = SDFail['SD']

                # Record filtering AS
                if SDFail['problemType'] == 'filtering':
                    ASSet = filteringAS
                else:
                    # Record dropping AS
                    ASSet = blockingAS

                if SDFail['failingASInfo'] is not None and type(SDFail['failingASInfo']) is dict:
                    failingASN = SDFail['failingASInfo']['asn']
                    if failingASN not in ASSet:
                        ASSet[SDFail['failingASInfo']['asn']] = {
                            'ASinfo': SDFail['failingASInfo'],
                            'exps': {expName: [failingSD]},
                        }
                    else:
                        if expName not in ASSet[failingASN]['exps']:
                            ASSet[failingASN]['exps'][expName] = [failingSD]
                        else:
                            ASSet[failingASN]['exps'][expName].append(failingSD)

    # Save general results
    utils.save_data_summary(results, SAVEDIRNAME, 'data_summary.txt')
    # Save analysed AS and blocking AS
    utils.save_ASInfos(results['seenAS'], results['analysedAS'], blockingAS, filteringAS, SAVEDIRNAME, 'seen_AS.txt', 'blocking_AS.txt', 'filtering_AS.txt', COMPRESSASINFOSFILE, ASTYPECACHE)
    # Save ASPath udp
    utils.save_ASPath(results['ASTraceByProto']['udp'], SAVEDIRNAME, 'AS_path_udp.txt')
    # Save ASPath tcp
    utils.save_ASPath(results['ASTraceByProto']['tcp'], SAVEDIRNAME, 'AS_path_tcp.txt')
    # Save IPs information
    utils.save_IPsInfos(IPSINFOS, SAVEDIRNAME, SEENIPSSAVEFILE, NOTMATCHIPSAVEFILE)
    # Saving ASInfoCache
    utils.save_ASCache(ASINFOSCACHE, ASINFOCACHEFILE)

    return 0
    
if __name__ == "__main__":
    main()
