"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import radix
import ipwhois
import os
import ipaddress
import ndjson
from scapy.all import *

"""
Load all the experiments to analyse
"""
def load_experiments(filePath):
    with open(filePath, 'r') as f:
        exps = ndjson.load(f)

    return exps

"""
Load AS caches
"""
def load_ASCaches(ASInfosCacheFilePath, ASTypeCacheFilePath, ASInfosCache, ASTypeCache):
    # Instanciate a radix tree for containing prefix
    ASInfosCache['tree'] = radix.Radix()

    # If cache file exist
    if os.path.exists(ASInfosCacheFilePath):
        with open(ASInfosCacheFilePath, 'r') as f:
            rawCache = ndjson.load(f)
            for entry in rawCache:
                if entry['asn_cidr'] not in ASInfosCache:
                    ASInfosCache['cache'][entry['asn_cidr']] = entry
                    ASInfosCache['tree'].add(entry['asn_cidr'])
    
    # If cache file exist
    if os.path.exists(ASTypeCacheFilePath):
        with open(ASTypeCacheFilePath, 'r') as f:
            rawCache = ndjson.load(f)
            for entry in rawCache:
                if entry['asn'] not in ASTypeCache:
                    ASTypeCache[entry['asn']] = entry['type']
    
    return

"""
Load host informations
"""
def load_hosts(hostFilePath):
    hosts = {}
    with open(hostFilePath, 'r') as f:
        for line in f:
            strippedLine = line.strip()
            if not strippedLine.startswith('#'):
                hostInfo = strippedLine.split()
                hosts[hostInfo[0].lower()] = {'ip':hostInfo[1]}

    return hosts

"""
Load destinations informations
"""
def load_destsInfos(destInfosFilePath):
    destsInfos = {}
    with open(destInfosFilePath, 'r') as f:
        for line in f:
            strippedLine = line.strip()
            if not strippedLine.startswith('#'):
                hostInfo = strippedLine.split()
                destsInfos[hostInfo[0]] = {
                    'originAS': str(hostInfo[1]),
                    'flowLabel': int(hostInfo[2]),
                    'cidr': str(hostInfo[3]),
                    'dstIP': str(hostInfo[4])
                }
    return destsInfos

"""
Retrieve all the destinations probed by the source
"""
def retrieve_src_destinations(fileName):
    try:
        probedDstsInfos = {}
        with open(fileName, 'r') as f:
            for line in f:
                strippedLine = line.strip()
                if not strippedLine.startswith('#'):
                    hostInfo = strippedLine.split()
                    probedDstsInfos[hostInfo[0]] = {
                        'dport': int(hostInfo[1]),
                        'dstIP': str(hostInfo[2]),
                        'flowLabel': int(hostInfo[3]),
                        
                    }

        return probedDstsInfos
    except FileNotFoundError:
        # If error, exit the script
        print('[ERROR] Cannot open file {}'.format(fileName))
        exit(-1)

"""
Filter packets to match source - destination
"""
def filter_pkts(srcIP, sndPkts, rcvPkts=[]):
    filterSndPkts = []
    filterRcvPkts = []

    # Filter sended pkts
    for pkt in sndPkts:
        # Filter by IPv6
        if pkt[IPv6].src == srcIP or pkt[IPv6].dst == srcIP:
            filterSndPkts.append(pkt)
    
    # Filter received pkts
    for pkt in rcvPkts:
        # Filter by IPv6
        if pkt[IPv6].src == srcIP:
            filterRcvPkts.append(pkt)

    return filterSndPkts, filterRcvPkts

"""
Compute AS trace
"""
def compute_ASTrace(dstASN, dstPrefix, hops, hopsCnt, ASInfoCache, IPsInfos):
    SDASTrace = {
        'ASTrace': {},
        'ASPath': [],
        'traceLen': hopsCnt,
        'seenAS': {}
    }

    # Retrieve all the AS of the experiment traceroute
    lastASNSeen = None
    for hopLimit, hopInfo in hops.items():
        infos = retrieve_ASInfo(hopInfo['ip'], ASInfoCache, IPsInfos)
        isDst = False
        IPv6Object = ipaddress.ip_address(hopInfo['ip'])
        networkObject = ipaddress.ip_network(dstPrefix)

        if infos is not None:
            # Check if dst ASN or dst prefix is reached
            if infos['asn'] == dstASN or (IPv6Object in networkObject):
                isDst = True
            
            # Insert in ASTrace
            SDASTrace['ASTrace'][hopLimit] = (infos, isDst)

            # Insert in ASPath
            if infos['asn'] != lastASNSeen:
                SDASTrace['ASPath'].append(infos['asn'])

            # Insert in seenAs
            if infos['asn'] not in SDASTrace['seenAS']:
                SDASTrace['seenAS'][infos['asn']] = infos
            
            lastASNSeen = infos['asn']
        else:
            # Check dst prefix is reached
            if IPv6Object in networkObject:
                isDst = True
            
            # Insert in ASTrace
            SDASTrace['ASTrace'][hopLimit] = (hopInfo['ip'], isDst)
    
    return SDASTrace

"""
Retrieve informations about a traceroute
"""
def retrieve_traceroute_infos(filename, expName):
    params = {}

    with open(filename, 'r') as f:
        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith(expName):
                splitted_line = line.split(',')
                params["START_TIME"] = splitted_line[1]
                params["END_TIME"] = splitted_line[2]

                return params
    
    return None

"""
Try to find a corresponding AS for a given IP 
"""
def retrieve_ASInfo(ipFailHop, ASInfoCache, IPsInfos):
    if ipFailHop is None or (ipFailHop in IPsInfos['notMatchIPs']):
        return None
    
    # Record seen IPs
    if ipFailHop not in IPsInfos['seenIPs']:
        IPsInfos['seenIPs'][ipFailHop] = 1
    else:
        IPsInfos['seenIPs'][ipFailHop] += 1

    # Find longest prefix
    longestPrefixNode = ASInfoCache['tree'].search_best(ipFailHop)

    # Match
    if longestPrefixNode is not None:
        if longestPrefixNode.prefixlen < 128:
            ASInfo = ASInfoCache['cache'][longestPrefixNode.prefix]
        else:
            ASInfo = ASInfoCache['cache'][longestPrefixNode.network]
        if ASInfo['asn'] != 'NA':
            return ASInfo
    else:
        # Try retrieving AS infos from ipwhois
        try:
            rawInfos = ipwhois.IPWhois(ipFailHop, timeout=2).lookup_whois(retry_count=2)
            if rawInfos is not None:
                filteredInfos = {
                    'asn': str(rawInfos['asn']),
                    'asn_cidr': str(rawInfos['asn_cidr']),
                    'asn_country_code': str(rawInfos['asn_country_code']),
                    'asn_date': str(rawInfos['asn_date']),
                    'asn_description': str(rawInfos['asn_description']),
                    'asn_registry': str(rawInfos['asn_registry']),
                }

                # Remove NA asn
                if filteredInfos['asn'] not in ['NA','None']:
                    # Add the new entry to the cache
                    if filteredInfos['asn_cidr'] not in ASInfoCache['cache']:
                        ASInfoCache['cache'][filteredInfos['asn_cidr']] = filteredInfos
                        ASInfoCache['tree'].add(filteredInfos['asn_cidr'])
                
                    return filteredInfos
        except:
            pass
    
    # Record not matched IP
    if ipFailHop not in IPsInfos['notMatchIPs']:
        IPsInfos['notMatchIPs'][ipFailHop] = True
    
    return None

"""
Try to determine the first filtering hop and the corresponding AS
"""
def find_firstFiltering_AS(SDExpTrace, ASTrace, ASInfosCache, IPsInfos):
    # Find the first filtering hop
    firstFiltHopHlim = None
    for hlim, hopInfo in SDExpTrace.items():
        if hopInfo['isFilter']:
            if firstFiltHopHlim is None or hlim < firstFiltHopHlim:
                firstFiltHopHlim = hlim

    # Determine corresponding AS
    if firstFiltHopHlim is None:
        return None
    ASInfos = retrieve_ASInfo(SDExpTrace[firstFiltHopHlim]['ip'], ASInfosCache, IPsInfos)
    
    # If None, search AS in ASTrace
    if ASInfos is None or type(ASInfos) is not dict:
        return ASTrace[firstFiltHopHlim][0] if firstFiltHopHlim in ASTrace else None
    else:
        return ASInfos

"""
Try to determine the dropping hop and the corresponding AS
"""
def determine_failingAS(SDExpTrace, ASTrace, ASInfoCache, IPsInfos):
    # Check if ASTrace is complete
    isComplete = False
    for hlim, hopInfo in ASTrace.items():
        if hopInfo[1] is True:
            isComplete = True
            break

    if not isComplete:
        return None

    hlimLastRespHop = max(SDExpTrace.keys()) if len(SDExpTrace.keys()) > 0 else 0
    
    # Handle case of no response from first hop of the traceroute
    if hlimLastRespHop == 0:
        # Return AS of first hop if info else return None
        return ASTrace[0][0] if 0 is ASTrace else None
    
    # Get last response type and determine the dropping hop
    msgTypeLastRespHop = SDExpTrace[hlimLastRespHop]['respType']
    hlimDropHop = hlimLastRespHop if msgTypeLastRespHop == 'ParamProblem' else hlimLastRespHop + 1

    # Retrieve ASN and IP for dropHop - 1, dropHop, and nextKnownHop
    # Based on the exp traceroute
    prevDropHopInfos = retrieve_ASInfo(SDExpTrace[hlimDropHop-1]['ip'], ASInfoCache, IPsInfos) if hlimDropHop-1 in SDExpTrace else None
    prevDropHopASN = prevDropHopInfos['asn'] if prevDropHopInfos is not None else None
    dropHopInfos = retrieve_ASInfo(SDExpTrace[hlimDropHop]['ip'], ASInfoCache, IPsInfos) if hlimDropHop in SDExpTrace else None
    dropHopASN = dropHopInfos['asn'] if dropHopInfos is not None else None

    # Based on the AS trace
    if prevDropHopASN is None:
        prevDropHopInfos = ASTrace[hlimDropHop-1][0] if hlimDropHop-1 in ASTrace else None
        prevDropHopASN = prevDropHopInfos['asn'] if prevDropHopInfos is not None and type(prevDropHopInfos) is dict else None
    if dropHopASN is None:
        dropHopInfos = ASTrace[hlimDropHop][0] if hlimDropHop in ASTrace else None
        dropHopASN = dropHopInfos['asn'] if dropHopInfos is not None and type(dropHopInfos) is dict else None
    
    # Find next responding hop
    if dropHopASN is None:
        hlimStartuZone = hlimDropHop
    else:
        hlimStartuZone = hlimDropHop + 1

    ASTraceHlimList = sorted(list(ASTrace))
    hlimNextRespHop = None

    for hlimHop in ASTraceHlimList[::-1]:
        if hlimNextRespHop is None:
            hlimNextRespHop = hlimHop
            continue
        if hlimStartuZone <= hlimHop and hlimHop < hlimNextRespHop:
            hlimNextRespHop = hlimHop

    nextRespHopInfos = ASTrace[hlimNextRespHop][0] if hlimNextRespHop in ASTrace and hlimNextRespHop is not None else None
    isNextRespHopDest = ASTrace[hlimNextRespHop][1] if hlimNextRespHop in ASTrace and hlimNextRespHop is not None else False
    nextRespHopASN = nextRespHopInfos['asn'] if nextRespHopInfos is not None and type(nextRespHopInfos) is dict else None
    
    if hlimDropHop == 1:
        if dropHopASN is None:
            return None
        else:
            if nextRespHopASN is None:
                return None
            else:
                return nextRespHopInfos

    # Check if prevDropHopASN or nextRespHopASN is None
    if prevDropHopASN is None or nextRespHopASN is None:
        return None

    # Compute length of uncertainty zone
    uZoneLen = hlimNextRespHop - (hlimDropHop)

    if uZoneLen < 0:
        return None

    if dropHopASN is not None:
        if uZoneLen - 1 <= 3:
            if prevDropHopASN == dropHopASN and dropHopASN != nextRespHopASN:
                if isNextRespHopDest: # Check if destination
                    return True
                else:
                    return nextRespHopInfos
            else:
                if isNextRespHopDest: # Check if destination
                    return True
                else:
                    return dropHopInfos
        else:
            return None
    else:
        # Check if uZone length is too big
        if uZoneLen > 3:
            return None # Uncertainty is too big
        else:
            if isNextRespHopDest:
                return True
            else:
                return nextRespHopInfos

"""
Save seen IPs information
"""
def save_IPsInfos(IPsInfos, savePath, filenameSeenIPs, filenameNotMatch):
    os.makedirs(savePath, exist_ok=True)

    # Record seen IPs
    with open("{}/{}".format(savePath, filenameSeenIPs), 'w') as f:
        f.write('#IP, seenCount\n')
        for IP, cnt in IPsInfos['seenIPs'].items():
            f.write('{}, {}\n'.format(IP, cnt))

    # Record not matched IPs
    with open("{}/{}".format(savePath, filenameNotMatch), 'w') as f:
        for IP in IPsInfos['notMatchIPs'].keys():
            f.write(IP + '\n')
    
    return

"""
Save results related to the traceroute analyses
"""
def save_results_traceAnalyse(data, savePath, srcIP, dstIP):
    os.makedirs(savePath, exist_ok=True)
    for expName, rslt in data.items():
        with open(savePath + 'trace_{}.txt'.format(expName), 'w') as f:
            f.write("-- Traceroute ({}) --\n\n".format(expName))
            f.write("SrcIP: {}\n".format(srcIP))
            f.write("DstIP: {}\n".format(dstIP))
            f.write("Max hlim sended: {}\n".format(rslt['maxHlimSeen']))
            f.write("Hops responding: {}\n\n".format(len(rslt['hops'])))
            f.write("**** Hops ****\n")
            for i in range(1, rslt['maxHlimSeen'] + 1):
                if i in rslt['hops']:
                    hopInfos = rslt['hops'][i]
                    codeType = ' - Code {}'.format(hopInfos['codeType']) if hopInfos['codeType'] is not None else ''
                    f.write("{} : {} ({}{}) {} ms {}\n".format(i, hopInfos['ip'], hopInfos['respType'], codeType, hopInfos['rtt'], "[filtered]" if hopInfos['isFilter'] else ""))
                else:
                    f.write("{} : ***\n".format(i))
    return

"""
Save detailed summary for the different source-destination
"""
def save_summary_test(testRslts, savePath, filename):
    os.makedirs(savePath, exist_ok=True)
    
    with open("{}/{}".format(savePath, filename), 'w') as f:
        f.write('\n# {} Experiments summary {}\n'.format(''.ljust(18, '*'), ''.ljust(18, '*')))
        for expName, expRslt in testRslts['dataByExp'].items():
            f.write('{} {} {}\n'.format(''.ljust(10, '-'), expName, ''.ljust(10, '-')))
            if len(expRslt['tracesFailure']) == 0:
                f.write('tracesFailure: []\n')
            else:
                f.write('tracesFailure: [\n\n')
                for fail in expRslt['tracesFailure']:
                    f.write('  {} {} {}\n'.format(''.ljust(7, '#'), fail['SD'], ''.ljust(7, '#')))
                    f.write('  problemType: {}'.format(fail['problemType']))
                    f.write('  failingASInfo: ')
                    print(fail['failingASInfo'] if fail['failingASInfo'] is not None else 'Unknown' , file=f)
                    f.write('  {}\n\n'.format(''.ljust(21, '#')))
                f.write(']\n')
            f.write('{}\n\n'.format(''.ljust(22+len(expName), '-')))

"""
Save the number of hops between source-destination in form of a matrix
"""
def save_HopsCntMatrix(hopsCntMatrix, expsName, savePath, filename):
    os.makedirs(savePath, exist_ok=True)
    
    with open("{}/{}".format(savePath, filename), 'w') as f:
        # Display hops count matrix
        f.write('# {} Nb of hops per Src-Dst {}\n'.format(''.ljust(19, '*'), ''.ljust(19, '*')))

        # Print column header
        f.write(''.ljust(35))
        for SDHopsCnt in hopsCntMatrix:
            f.write('{}  '.format(SDHopsCnt['SD']))
        f.write('\n')
        
        # Print row
        for expName in expsName:
            f.write('{}{}'.format(expName, ''.ljust(35 - len(expName))))
            for SDHopsCnt in hopsCntMatrix:
                f.write('{}{}'.format(SDHopsCnt['hopsCntByExp'][expName], ''.ljust(8-len(str(SDHopsCnt['hopsCntByExp'][expName])))))
            f.write('\n')
        
        f.write('\n\n')

"""
Save global summary for all source-destination
"""
def save_data_summary(data, savePath, filename):
    os.makedirs(savePath, exist_ok=True)

    with open("{}/{}".format(savePath, filename), 'w') as f:
        f.write('# {} Experiments summary {}\n'.format(''.ljust(14, '*'), ''.ljust(14, '*')))
        f.write('----- General statistics -----\n')
        f.write('Nb of tested SD: {}\n'.format(data['nbSDTested']))
        f.write('Nb of NoExtHdr trace. reaching dst (UDP): {}/{} ({:.1f}%)\n'.format(data['nbSDUDPReached'], data['nbSDTested'], float((data['nbSDUDPReached']/data['nbSDTested'])*100)))
        f.write('Nb of NoExtHdr trace. reaching dst (TCP): {}/{} ({:.1f}%)\n'.format(data['nbSDTCPReached'], data['nbSDTested'], float((data['nbSDTCPReached']/data['nbSDTested'])*100)))
        f.write('------------------------------\n')
        f.write('{}{}\n'.format(''.ljust(35), 'SUCCESS/TOTAL (FORW/BLOCK/FILT/TOTAL) trace.'))
        for expName, expData in data['dataByExp'].items():
            percent = (expData['nbSDForwarding'] / expData['nbSDTotal']) * 100
            f.write('{}{}'.format(expName, ''.ljust(35 - len(expName))))
            f.write('{:.1f}% ({}/{}/{}/{})\n'.format(float(percent), expData['nbSDForwarding'], expData['nbSDBlocking'], expData['nbSDFiltering'], expData['nbSDTotal']))
    return

"""
Save the analysed AS, blocking AS, and filtering AS
"""
def save_ASInfos(seenAS, analysedAS, blockingAS, filteringAS, savePath, filenameSeenAS, filenameBlockingAS, filenameFilteringAS, filenameAnalysedAS, ASTypeCache):
    os.makedirs(savePath, exist_ok=True)

    # Save analysed AS
    with open("{}/{}".format(savePath, filenameSeenAS), 'w') as f:
        f.write('{} Seen AS {}\n\n'.format(''.ljust(14, '*'), ''.ljust(14, '*')))
        f.write('General infos:\n')
        f.write('  Number of seen AS: {}\n\n'.format(len(seenAS)))
        f.write('Details:\n')
        for asn, ASinfo in seenAS.items():
            f.write('{}: {}\n'.format(str(asn), ASinfo))
    
    # Save blocking AS
    with open("{}/{}".format(savePath, filenameBlockingAS), 'w') as f:
        f.write('{} Blocking AS {}\n\n'.format(''.ljust(14, '*'), ''.ljust(14, '*')))
        f.write('General infos:\n')
        f.write('  NbOfBlockingAS: {}\n\n'.format(len(blockingAS)))
        f.write('Details:\n')
        for asn, details in blockingAS.items():
            f.write('{}:\n'.format(str(asn)))
            f.write('  asn: {}\n'.format(str(details['ASinfo']['asn'])))
            f.write('  asn_cidr: {}\n'.format(details['ASinfo']['asn_cidr']))
            f.write('  asn_country_code: {}\n'.format(details['ASinfo']['asn_country_code']))
            f.write('  asn_date: {}\n'.format(details['ASinfo']['asn_date']))
            f.write('  asn_description: {}\n'.format(details['ASinfo']['asn_description']))
            f.write('  asn_registry: {}\n'.format(details['ASinfo']['asn_registry']))
            f.write('  blockedExps: [\n')
            for expName, failingSDs in details['exps'].items():
                f.write('    {} ('.format(expName))
                for SD in failingSDs:
                    f.write(SD)
                    if SD != failingSDs[-1]:
                        f.write(', ')
                f.write(')\n')
            f.write('  ]\n\n')

    # Save filtering AS
    with open("{}/{}".format(savePath, filenameFilteringAS), 'w') as f:
        f.write('{} Filtering AS {}\n\n'.format(''.ljust(14, '*'), ''.ljust(14, '*')))
        f.write('General infos:\n')
        f.write('  NbOfFilteringAS: {}\n\n'.format(len(filteringAS)))
        f.write('Details:\n')
        for asn, details in filteringAS.items():
            f.write('{}:\n'.format(str(asn)))
            f.write('  asn: {}\n'.format(str(details['ASinfo']['asn'])))
            f.write('  asn_cidr: {}\n'.format(details['ASinfo']['asn_cidr']))
            f.write('  asn_country_code: {}\n'.format(details['ASinfo']['asn_country_code']))
            f.write('  asn_date: {}\n'.format(details['ASinfo']['asn_date']))
            f.write('  asn_description: {}\n'.format(details['ASinfo']['asn_description']))
            f.write('  asn_registry: {}\n'.format(details['ASinfo']['asn_registry']))
            f.write('  filteredExps: [\n')
            for expName, failingSDs in details['exps'].items():
                f.write('    {} ('.format(expName))
                for SD in failingSDs:
                    f.write(SD)
                    if SD != failingSDs[-1]:
                        f.write(', ')
                f.write(')\n')
            f.write('  ]\n\n')

    # Save a compressed version of the information
    with open("{}/{}".format(savePath, filenameAnalysedAS), 'w') as f:
        f.write("#asn; asn_description; isFiltering; isBlocking; type\n")
        for asn, desc in analysedAS.items():
            # Find AS type
            ASType = ASTypeCache[int(asn)] if int(asn) in ASTypeCache else 'noEntry'

            isFiltering = False
            isBlocking = False
            # If filtering AS 
            if asn in filteringAS:
                isFiltering = True

            # If blocking AS
            if asn in blockingAS:
                isBlocking = True

            f.write("{}; {}; {}; {}; {}\n".format(asn, desc, isFiltering, isBlocking, ASType))
    return

"""
Save AS path for each source-destination
"""
def save_ASPath(ASPathBySD, savePath, filename):
    os.makedirs(savePath, exist_ok=True)

    # Save AS path
    with open("{}/{}".format(savePath, filename), 'w') as f:
        f.write('{} ASPath by source-destination {}\n\n'.format(''.ljust(14, '*'), ''.ljust(14, '*')))
        for SDName, pathData in ASPathBySD.items():
            f.write('{} {} {}\n\n'.format(''.ljust(6, '-'), SDName, ''.ljust(6, '-')))
            f.write('AS traceroute:\n')
            for i in range(1, pathData['traceLen']+1):
                if i in pathData['ASTrace']:
                    isDst = '[Dest]' if pathData['ASTrace'][i][1] is True else ''
                    if type(pathData['ASTrace'][i][0]) is dict:
                        f.write(" {} : {} ({}) {}\n".format(i, pathData['ASTrace'][i][0]['asn'], pathData['ASTrace'][i][0]['asn_description'], isDst))
                    else:
                        f.write(" {} : {} ({}) {}\n".format(i, pathData['ASTrace'][i][0], "AS not found", isDst))
                else:
                    f.write(" {} : ***\n".format(i))
            f.write('\n')

            # AS path
            f.write('AS path:\n')
            for i in range(0, len(pathData['ASPath'])):
                f.write(" {} : {}\n".format(i+1, pathData['ASPath'][i]))
                
            f.write('{}\n\n'.format(''.ljust(12+len(SDName), '-')))

"""
Save AS cache
"""
def save_ASCache(ASInfoCache, savePath):
    with open(savePath, 'w') as f:
        writer = ndjson.writer(f, ensure_ascii=True)
        for network, ASInfo in ASInfoCache['cache'].items():
            writer.writerow(ASInfo)
    
    return
