"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import config as config

from scapy.all import *
from ndjson import load
from modules.traceroute import Traceroute
from modules.forge import Forge

class Experiments:
    """ Experiments class
    This class is responsible of performing all the experiments towards a 
    destination.

    Attributes
    ----------
    path : str
        Save path

    name : str 
        The destination attributed name
    
    dstInfos : list
        Contain the IPv6 address, dport, flowlabel related to the destination
    
    nbHopsByProto : dict
        Contain the number of hops by protocol to between the source and the
        destination
    
    protocols : list
        Lists of protocol to test
    
    expsToSkip : list
        List of experience name to skip
    
    expsInputFilename : str
        The input name containing the configuration of experiments that need
        to be performed

    results : dict
        Contain the results of the various experiments 
    """

    def __init__(self, rootPath, name, dstInfos, nbHopsByProto, protocols, expsToSkip, expsInputFilename):
        self.path = rootPath + '/' + name
        self.name = name
        self.dstInfos = dstInfos
        self.nbHopsByProto = nbHopsByProto
        self.protocols = protocols
        self.expsToSkip = expsToSkip
        self.expsInputFilename = expsInputFilename
        self.results = {}

    """
    Load all experiments that need to be performed
    """
    def load_experiments(self):
        with open(self.expsInputFilename) as f:
            exps = load(f)
        
        return exps
    
    """
    Write experiences results into files
    """
    def write_results(self):
        tcpFileName = self.path + '/tcp/experiments_result.txt'
        udpFileName = self.path + '/udp/experiments_result.txt'
        noL4FileName = self.path + '/noL4/experiments_result.txt'

        os.makedirs(self.path+'/tcp/', exist_ok=True)
        os.makedirs(self.path+'/udp/', exist_ok=True)
        os.makedirs(self.path+'/noL4/', exist_ok=True)
        try:
            with open(tcpFileName, 'a') as f1, open(udpFileName, 'a') as f2, open(noL4FileName, 'a') as f3:
                f1.write('#exp_name,protocol,nbPktsSnd,startTime,endTime\n')
                f2.write('#exp_name,protocol,nbPktsSnd,startTime,endTime\n')
                f3.write('#exp_name,protocol,nbPktsSnd,startTime,endTime\n')

                for exp_name, results in self.results.items():
                    # Write tcp results
                    if 'tcp' in results:
                        f1.write(exp_name + ',tcp,' + str(config.PACKET_TO_SEND) + ',')
                        f1.write(str(results['tcp'][0]) + ',' + str(results['tcp'][1]) + '\n')
                        f2.flush()
                    
                    # Write udp results
                    if 'udp' in results:
                        f2.write(exp_name + ',udp,' + str(config.PACKET_TO_SEND) + ',')
                        f2.write(str(results['udp'][0]) + ',' + str(results['udp'][1]) + '\n')
                        f2.flush()
                    
                    # Write noL4 results
                    if 'noL4' in results:
                        f3.write(exp_name + ',noL4,' + str(config.PACKET_TO_SEND) + ',')
                        f3.write(str(results['noL4'][0]) + ',' + str(results['noL4'][1]) + '\n')
                        f3.flush()
        except IOError:
            print('[ERROR] Not able to write the experiments results of {}'.format(self.name))

    """
    Send a number of experiment packets (packets containing an extension header)
    with a fixed hoplimit.
    """
    def send_packets_EH(self, protocol, expConfig):
        # Initialize the probe generator
        forgeGen = Forge(self.dstInfos, protocol, expConfig)

        # Start time
        startTime = time.time()

        # Send experiment packets
        for i in range(0, config.PACKET_TO_SEND):
            probe = forgeGen.forge(config.PROBE_ID, 255)
            if (probe == -1):
                return -1

            send(probe, verbose=0)

            if i != config.PACKET_TO_SEND-1:
                time.sleep(config.SEND_TIMEOUT)

            config.PROBE_ID += 1                
            
        # Stop time
        endTime = time.time()
        
        if config.VERBOSE: print('   {} [{}] to {}: Sending {} packets --> COMPLETE'.format(expConfig['name'], protocol, self.name, str(config.PACKET_TO_SEND)))
        
        # Save results
        self.results[expConfig['name']][protocol] = [startTime, endTime]

        return 0

    """
    Perform the methodology to test an extension header
    (Perform traceroutes, send experiment packets with fixed hop limit)
    """
    def perform_experiment(self, protocol, expConfig, nbHopsToDst):
        # Check if experiment must be ignored
        for expToSkip in self.expsToSkip:
            expName = expConfig['name'] + protocol
            if expToSkip.lower() in expName.lower():
                return -1
        
        # If HBH discard experiment, slow down rate of traceroute
        if config.SLOW_HBHDISCARD:
            slowMatches = ['HbH', 'discard']
            isSlow = True if all(x in expConfig['name'] for x in slowMatches) else False
        else:
            isSlow = False

        # STEP 1: Perform a traceroute with packet containing the EH
        rtn = Traceroute(self.dstInfos, protocol, expConfig['name'], expConfig, False, isSlow=isSlow, maxHops=nbHopsToDst, saveDirPath='{}/{}/'.format(self.path, protocol)).launch()
        if (rtn != 0):
            print('   {} [{}] to {}: Traceroute with EH --> ERROR'.format(expConfig['name'], protocol, self.name))
        else:
            if config.VERBOSE: print('   {} [{}] to {}: Traceroute with EH --> COMPLETE'.format(expConfig['name'], protocol, self.name))
        
        if config.EXP_PROBING:
            # Wait before performing the next step
            time.sleep(config.WAIT_STEP)

            # STEP 2: Send packets to the destination address with the EH
            rtn = self.send_packets_EH(protocol, expConfig)
            if (rtn != 0):
                if config.VERBOSE: print('   {} [{}] to {}: ERROR'.format(expConfig['name'], protocol, self.name))
        
        return 0

    """
    Perform all the experiments for a destination
    """
    def start_experiments(self):
        # Load all the experiments to perform
        exps = self.load_experiments()

        # Iterate through the experiments
        for exp in exps:
            self.results[exp['name']] = {}

            # Perform the experiment with a given protocol
            if exp['name'] in ['noNextHdr', 'ethernet']:
                rtn = self.perform_experiment('noL4', exp, self.nbHopsByProto[self.protocols[0]])
            elif exp['name'] in ['esp']:
                rtn = self.perform_experiment('udp', exp, self.nbHopsByProto[self.protocols[0]])
            else:
                for protocol in self.protocols:
                    rtn = self.perform_experiment(protocol, exp, self.nbHopsByProto[protocol])
                    if rtn == 0:
                        time.sleep(config.PROTO_INTERVAL)

            # Wait before perfoming the next
            if exp != exps[-1] and rtn == 0:
                time.sleep(config.EXP_INTERVAL)

        # Write time
        if config.EXP_PROBING:
            self.write_results()
    
