"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import config as config
from scapy.all import IPv6
from scapy.all import TCP
from scapy.all import UDP
from scapy.all import IPv6ExtHdrDestOpt
from scapy.all import IPv6ExtHdrFragment
from scapy.all import IPv6ExtHdrHopByHop
from scapy.all import IPv6ExtHdrRouting
from scapy.all import IPv6ExtHdrSegmentRouting
from scapy.all import Jumbo
from scapy.all import HAO
from scapy.all import Pad1
from scapy.all import PadN
from scapy.all import RouterAlert
from scapy.all import SecurityAssociation
from scapy.all import AH
from scapy.all import ESP
from scapy.all import struct
from scapy.all import Raw
from scapy.all import Ether
from scapy.all import RandMAC
from scapy.volatile import RandNum

PAYLOADDATA = config.PAYLOAD_DATA
SPLITBYTE = b'\x00'

class Forge:
    """ Forge class
    This class is responsible of forging probes containing EH.

    Attributes
    ----------
    dstInfos : list
        Contain the IPv6 address, dport, flowlabel related to the destination

    protocol : str
        Provide the transport layer to need to insert in the probe
    
    probeConfig : dict
        Provide more details on the EH that need to be inserted in the probe
    """

    def __init__(self, dstInfos, protocol, probeConfig):
        self.dstInfos = dstInfos
        self.protocol = protocol
        self.probeConfig = probeConfig

    """
    Create a destination extension header

    To use:
        - Add the string 'dstHdr' in extHdr of you input ndjson job file
        - Add a entry 'dstHdrConfig' containing a dictionnary
        - If you want options, add a entry 'option' contaning the name of 
        the option you want

    Available options:
        - HAO
        - Pad1
        - PadN --> To configure the PadN, insert an entry "PadN" in dstHdrConfig
                   containing a dictionnary with the possible entry:
                   - "action": Specify the opt action (skip, discard)
                   - "type": Specify the type of PadN (unkw, real)
                   - "padding": Specify the length of the padding to insert
    """
    def create_dest_extHdr(self, config):
        # Create IPv6 destination header
        header = IPv6ExtHdrDestOpt()

        if config == None:
            return header
        
        # Add options
        if 'option' in config:
            option = config['option']
            if option == 'HAO':
                header.options += HAO()
            elif option == 'Pad1':
                header.options += Pad1()
            elif option == 'PadN':
                # Check if configuration
                if option not in config:
                    print("[ERROR] No PadN configuration detected")
                    return None
                
                # Determine otype
                padType = config['PadN']['type']
                padding = config['PadN']['padding']

                # Check if padding is not 0
                if padding > 0:
                    if config['PadN']['action'] == 'skip':
                        if padType == 'unkw':
                            otype = 22 # act:00 c:0 typ:10110
                        else:
                            otype = 1 # act:00 c:0 typ:00001
                    elif config['PadN']['action'] == 'discard':
                        if padType == 'unkw':
                            otype = 150 # act:10 c:0 typ:10110
                        else:
                            otype = 129 # act:10 c:0 typ:00001

                    else:
                        print("[ERROR] Unknow opt action")
                        return None
                    
                    # Add the padding
                    if padType == 'unkw' or padType == 'real':
                        if padding <= 259:
                            header.options += PadN(otype=otype, optdata=(padding-2) * '\x00')
                        else:
                            nbOptions = int((padding-2) / 256)
                            for i in range(0, nbOptions):
                                header.options += PadN(otype=otype, optdata= 254 * '\x00')
                            header.options += PadN(otype=otype, optdata=(padding - (256*nbOptions) - 2) * '\x00')
                    else:
                        print('[ERROR] Unknow type of PadN')
                        return None
            else:
                print("[ERROR] Unkown option: " + option)
                return None
    
        return header

    """
    Create a hop-by-hop extension header

    To use:
        - Add string the 'HbHHdr' in extHdr of you input ndjson job file
        - Add a entry 'HbHHdrConfig' containing a dictionnary
        - If you want options, add a entry 'option' contaning the name of 
        the option you want

    Avaivable options:
        - Jumbo
        - HAO
        - RoutingAlert
        - Pad1
        - PadN --> To configure the PadN, insert an entry "PadN" in HbHHdrConfig
                   containing a dictionnary with the possible entry:
                   - "action": Specify the opt action (skip, discard)
                   - "type": Specify the type of PadN (unkw, real)
                   - "padding": Specify the length of the padding to insert
    """
    def create_hopByhop_extHdr(self, config):
        # Create IPv6 hop by hop header
        header = IPv6ExtHdrHopByHop()

        if config == None:
            return header
        
        # Add options
        if 'option' in config:
            option = config['option']
            if option == 'Jumbo':
                header.options += Jumbo()
            elif option == 'HAO':
                header.options += HAO()
            elif option == 'RouterAlert':
                header.options += RouterAlert()
            elif option == 'Pad1':
                header.options += Pad1()
            elif option == 'PadN':
                # Check if configuration
                if option not in config:
                    print("[ERROR] No PadN configuration detected")
                    return None
                
                # Determine otype
                padType = config['PadN']['type']
                padding = config['PadN']['padding']

                # Check if padding is not 0
                if padding > 0:
                    if config['PadN']['action'] == 'skip':
                        if padType == 'unkw':
                            otype = 54 # act:00 c:1 typ:10110
                        else:
                            otype = 1 # act:00 c:0 typ:00001
                    elif config['PadN']['action'] == 'discard':
                        if padType == 'unkw':
                            otype = 182 # act:10 c:1 typ:10110
                        else:
                            otype = 129 # act:10 c:0 typ:00001
                    else:
                        print("[ERROR] Unknow opt action")
                        return None
                    
                    # Add the padding
                    if padType == 'unkw' or padType == 'real':
                        if padding <= 259:
                            header.options += PadN(otype=otype, optdata=(padding-2) * '\x00')
                        else:
                            nbOptions = int((padding-2) / 256)
                            for i in range(0, nbOptions):
                                header.options += PadN(otype=otype, optdata= 254 * '\x00')
                            header.options += PadN(otype=otype, optdata=(padding - (256*nbOptions) - 2) * '\x00')
                    else:
                        print('[ERROR] Unknow type of PadN')
                        return None
            else:
                print("[ERROR] Unkown option: " + option)
                return None

        return header
    
    """
    Create probe with fragmentation extension header

    To use:
        - Add the string 'fragHdr' in extHdr of you input ndjson job file
        - Add a entry 'fragHdrConfig' containing a dictionnary

    Available config:
        - size -> [int] Size of the fragments
        - m -> Set the MoreFragment flag value (0 or 1)
    """
    def create_frag_probe(self, config, l3, l4, data):
        if config == None:
            return None
        
        # Create and configure the header
        header = IPv6ExtHdrFragment()
        header.id = RandNum(min=0, max=4294967296)
        
        if 'm' in config:
            header.m = config['m']
        else:
            return None
        
        # Compute dataLen
        size = config['size']
        if self.protocol == 'udp':
            newDataSize = size - 70
        elif self.protocol == 'tcp':
            newDataSize = size - 82
        else:
            return None

        # Insert the right number of padding to reach the correct size
        paddingLen = newDataSize - len(data[Raw].load)
        newData = Raw(load=data[Raw].load + str.encode((paddingLen * '\x00')))

        # Combine all the layers
        return l3/header/l4/newData

    """
    Create a routing extension header

    To use:
        - Add string 'routingHdr' in extHdr of you input ndjson job file
        - Add a entry 'routingHdrConfig' containing a dictionnary

    Available config:
        - type        --> [Int] Routing type btw 1 and 6.
        - segLeft     --> [Int] Number of segments left
        - addresses   --> [Array] List of IPv6 addresses
    """
    def create_routing_extHdr(self, config):
        # Create IPv6 routing header
        header = IPv6ExtHdrRouting()

        if config == None:
            return None

        if 'type' in config:
            if config['type'] == 4:
                header = IPv6ExtHdrSegmentRouting()
                header.lastentry = 0
            elif config['type'] == 5:
                header.type = 5
                header.reserved = 0x00080000
            elif config['type'] == 6:
                header.type = 6
                header.reserved = 0x00000008
            else:
                header.type = config['type']

            if 'segLeft' in config:
                header.segleft = config['segLeft']
            else:
                header.segleft = 0

            if 'addresses' in config:
                if (len(config['addresses']) == 0 and config['type'] != 5 and config['type'] != 6):
                    header.addresses = [self.dstInfos[0]]
                else:
                    for addr in config['addresses']:
                        header.addresses.append(addr)
        else:
            print('[ERROR] Unknow routing type')
            return None

        return header
    
    """
    Create an authentication header

    To use:
        - Add string 'ahHdr' in extHdr of you input ndjson job file
        - Add a entry 'ahHdrConfig' containing a dictionnary

    Available config:
        - spi        --> [String(hex)] Security parameter index
        - authAlgo   --> [String] Auth algorithm ('HMAC-SHA1-96', 'SHA2-256-128', 'SHA2-384-192', 
                                'SHA2-512-256', 'HMAC-MD5-96', 'AES-CMAC-96')
        - authKey    --> [String] Auth key 
    """
    def create_auth_extHdr(self, config, probe):
        if config is None:
            return None
        
        # Create the security association and authenticate the probe
        if 'spi' in config and 'authAlgo' in config and 'authKey' in config:
            sa = SecurityAssociation(proto=AH, spi=int(config['spi'], 16), auth_algo=config['authAlgo'], auth_key=str.encode(config['authKey']))
            return sa.encrypt(probe)
        else:
            print('[ERROR] No valid configuration')
            return None

    """
    Create an ESP header

    To use:
        - Add string 'espHdr' in extHdr of you input ndjson job
        - Add a entry 'espHdrConfig' containing a dictionnary

    Available config:
        - spi        --> [String(hex)] Security parameter index (MUST be greater than 255 - See RFC4303 for more info)
        - cryptAlgo   --> [String] Encryption algorithm ('3DES', 'Blowfish', 'AES-GCM', 
                                'DES', 'AES-CBC', 'CAST', 'AES-CTR', 'NULL')
        - cryptKey    --> [String] Encryption key
    """
    def create_esp_extHdr(self, config, probe):
        if config is None:
            return None

        # Create the security association and authenticate the probe
        if 'spi' in config and 'cryptAlgo' in config and 'cryptKey' in config:
            sa = SecurityAssociation(proto=ESP, spi=int(config['spi'], 16), crypt_algo=config['cryptAlgo'], crypt_key=str.encode(config['cryptKey']))
            return sa.encrypt(probe)
        else:
            print('[ERROR] No valid configuration')
            return None
    
    """
    Add the extention header and combine the multiple layer of the probe
    """
    def probe_assembly(self, l3, l4, data, probeID):
        extHdrName = self.probeConfig['extHdr']

        if extHdrName == 'raw':
            return l3/l4/data
        elif extHdrName == 'dstHdr':
            header = self.create_dest_extHdr(self.probeConfig['dstHdrConfig'])
        elif extHdrName == 'HbHHdr':
            header = self.create_hopByhop_extHdr(self.probeConfig['HbHHdrConfig'])
        elif extHdrName == 'routingHdr':
            header = self.create_routing_extHdr(self.probeConfig['routingHdrConfig'])
        elif extHdrName == 'noNextHdr':
            # Modify the next header field
            l3[IPv6].nh = 59
            return l3/data
        elif extHdrName == 'ethernet':
            # Modify the next header field
            l3[IPv6].nh = 143
            # Create ethernet header
            ether = Ether(src=RandMAC(), dst=RandMAC(), type=0x86DD)
            # Create IPv6 encapsulation
            encap = IPv6(dst=self.dstInfos[0], hlim=50, fl=0) / UDP(sport=self.probeConfig['sp'], dport=self.probeConfig['sp'])
            
            return l3/ether/encap/data
        elif extHdrName == 'fragHdr':
            probe = self.create_frag_probe(self.probeConfig['fragHdrConfig'], l3, l4, data)
            if probe is None:
                return -1
            else:
                return probe 
        elif extHdrName == 'ahHdr':
            probe = l3/l4/data

            # Authenticate the probe
            authProbe = self.create_auth_extHdr(self.probeConfig['ahHdrConfig'], probe)
            if authProbe is None:
                return -1
            else:
                return authProbe
        elif extHdrName == 'espHdr':
            probe = l3/l4/data

            # Encrypt the probe
            encryptedProbe = self.create_esp_extHdr(self.probeConfig['espHdrConfig'], probe)
            if encryptedProbe is None:
                return -1
            else:
                encryptedProbe.seq = probeID
                return encryptedProbe
        else:
            print("[ERROR] Unknow extension header type")
            return -1
        
        # Combine all the layers
        if header is None:
            return -1
        else:
            return l3/header/l4/data

    """
    Forge the probe
    """
    def forge(self, probeID, hlim):
        # Creation of the IPv6 header
        l3 = IPv6()
        l3.hlim = hlim
        l3.dst = self.dstInfos[0]
        l3.fl = self.dstInfos[2]

        # Creation of the layer 4 header
        if self.protocol == 'udp':
            l4 = UDP(sport=self.probeConfig['sp'], dport=self.dstInfos[1])
        elif self.protocol == 'tcp':
            l4 = TCP(sport=self.probeConfig['sp'], dport=self.dstInfos[1], flags='S', seq=probeID)
        elif self.protocol == 'noL4':
            l4 = None
        else:
            print('[ERROR] Unknow layer 4 protocol')
            return -1
        
        # Creation of the probe payload
        data = Raw(load=PAYLOADDATA + str.encode(str(probeID)) + b'\x00')
        
        if 'extHdr' not in self.probeConfig:
            return l3/l4/data
        else:
            return self.probe_assembly(l3, l4, data, probeID)
