"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import time
import config as config

from scapy.all import *
from modules.forge import Forge
from modules.forge import PAYLOADDATA
from modules.forge import SPLITBYTE

class Traceroute:
    """ Traceroute class
    This class is responsible of performing traceroutes contining EH.

    Attributes
    ----------
    dstInfos : list
        Contain the IPv6 address, dport, flowlabel related to the destination
    
    protocol : str
        Provide the transport layer to need to insert in the probe
    
    expName : str
        The name of the experiment
    
    expConfig : dict
        Contain the configuration of the EH 
    
    isCountHops : bool
        Determine if the number of hops need to count or not
    
    isSlow : bool
        Determine if the traceroute need to be slowed down for a experiment
    
    maxHops : int
        The number of the maximal hop limit to send
    
    saveDirPath : str
        Path towards the saving directory
    
    saveFilename : str
        The name of the saving file

    """

    def __init__(self, dstInfos, protocol, expName, expConfig, isCountHops, isSlow=False, maxHops=config.DEFAULT_MAXHOP, saveDirPath=None, saveFilename=config.DEFAULT_SAVEFILENAME):
        self.dstInfos = dstInfos
        self.protocol = protocol
        self.expName = expName
        self.expConfig = expConfig
        self.isCountHops = isCountHops
        self.isSlow = isSlow
        self.maxHops = maxHops if maxHops == config.DEFAULT_MAXHOP else maxHops + config.ADD_HOPS
        self.saveDirPath = saveDirPath
        self.saveFilename = saveFilename
        
    """
    Generate a list containing all traceroute probes
    """
    def constructs_traceroute_pkts(self):
        tracePkts = []
        sendedIDs = {}

        # Initialize the probe generator
        forgeGen = Forge(self.dstInfos, self.protocol, self.expConfig)

        for hlim in range(1, self.maxHops+1):
            probe = forgeGen.forge(config.PROBE_ID, hlim)
            if (probe == -1):
                return -1
            
            tracePkts.append(probe)
            sendedIDs[config.PROBE_ID] = hlim

            config.PROBE_ID += 1

        return tracePkts, sendedIDs

    """
    Perform a traceroute and count the number of hops
    """
    def countHops_traceroute(self, tracePkts, sendedIDs):
        inter = config.SLOW_TRACE_TIMEOUT if self.isSlow else 0
        ans, unans = sr(tracePkts, timeout=config.WAIT_RESP, verbose=False, inter=inter)

        hlimDst = None
        try:
            for s, r in ans:
                # Check if response from destination
                if r.haslayer(IPv6):
                    if r[IPv6].src == self.dstInfos[0]:
                        if r.haslayer(Raw):
                            respID = int(r[Raw].load.partition(PAYLOADDATA)[2].partition(SPLITBYTE)[0])
                            if r.haslayer(ICMPv6ParamProblem):
                                if hlimDst is None or sendedIDs[respID] < hlimDst:
                                    hlimDst = sendedIDs[respID]
                            elif r.haslayer(ICMPv6DestUnreach):
                                if hlimDst is None or sendedIDs[respID] < hlimDst:
                                    hlimDst = sendedIDs[respID]
                        elif r.haslayer(TCP):
                            if 'SA' == r[TCP].flags:
                                respID = r[TCP].ack - 1
                                if hlimDst is None or sendedIDs[respID] < hlimDst:
                                    hlimDst = sendedIDs[respID]
        except:
            return -1
            
        # Return error if no response from destination
        if hlimDst is None:
            return -1
        else:
            return hlimDst

    """
    Save informations related to the performed traceroute
    """
    def save_informations(self, starTime, endTime):
        if self.saveDirPath is None:
            return
        
        # Check if storage file exists
        filePath = self.saveDirPath + self.saveFilename
        if not os.path.exists(filePath):
            os.makedirs(self.saveDirPath, exist_ok=True)
            with open(filePath, 'w') as f:
                f.write("#expName,starTime,endTime\n")
                f.write("{},{},{}\n".format(self.expName,str(starTime),str(endTime)))
                f.flush()
        else:
            with open(filePath, 'a') as f:
                f.write("{},{},{}\n".format(self.expName,str(starTime),str(endTime)))
                f.flush()
        
        return

    """
    Launch the traceroute
    """
    def launch(self):
        if self.expConfig is None:
            return -1
        
        nbHops = None

        # Record start time
        startTime = time.time()

        # Send probes
        for attempt in range(0, config.MAX_ATTEMPS):
            # Construct the list of pkts to send
            tracePkts, sendedIDs = self.constructs_traceroute_pkts()

            if self.isCountHops:
                hlimDst = self.countHops_traceroute(tracePkts, sendedIDs)
                if hlimDst != -1:
                    if nbHops is None or hlimDst > nbHops:
                        nbHops = hlimDst
            else:
                inter = config.SLOW_TRACE_TIMEOUT if self.isSlow else 0
                send(tracePkts, verbose=False, inter=inter)
                nbHops = 0
            
            time.sleep(config.WAIT_RESP)
        
        # Record end time       
        endTime = time.time()

        # Save informations
        self.save_informations(startTime, endTime)
        
        if nbHops is None:
            return config.DEFAULT_MAXHOP
        else:
            return nbHops