"""
Copyright 2022 Raphaël Léas
University of Liège

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import os
import argparse
import config as config

from modules.capture import Capture
from modules.experiments import Experiments
from modules.traceroute import Traceroute

"""
Parse program arguments
"""
def parseArgs():
    parser = argparse.ArgumentParser()

    # Interface to listen to
    parser.add_argument('--i', type=str, required=True, help='Interface to listen')

    # Experiments to ignore
    parser.add_argument('--s', '--skip', nargs='+', default=[], help="Exp to skip")

    args = parser.parse_args()
    return args

"""
Read the destination addresses and their corresponding information from a file
"""
def retrieve_destinations(fileName):
    dstAddress = {}
    
    # Open the file
    try:
        file = open(fileName, 'r')

        # Record the data
        for line in file:
            strippedLine = line.strip()
            data = strippedLine.split()
            if data[0] != '#':
                dstAddress[data[0]] = [data[2], int(data[1]), int(data[3])]
        
        # Close the file
        file.close()

        return dstAddress
    except FileNotFoundError:
        # If error, exit the script
        print('[ERROR] Cannot open file {}'.format(fileName))
        exit()

"""
Create directory structure of the results
"""
def setup_dir_structure():
    # Create the results directory
    if (os.path.isdir(config.TESTS_DIR) is False):
        os.mkdir(config.TESTS_DIR)
    
    # Create capture dir
    if (os.path.isdir(config.TESTS_DIR + '/captures') is False):
        os.mkdir(config.TESTS_DIR + '/captures')

"""
Perform traceroute towards a destination
"""
def traceroute_to_destination(id, name, dstInfos, maxHops=None):
    # Configure the pkt
    pktConfig = {"name":"NoHeader","sp":33434, "extHdr":"raw"}

    # For tcp and udp
    nbHopsByProto = {}
    for proto in config.PROTOCOLS:
        if config.VERBOSE: print('   Using {} --> Started'.format(proto))
        path = config.TESTS_DIR + '/' + name + '/' + proto + '/'

        # Launch the traceroute
        if config.REDUCE_HOPS:
            if maxHops is None:
                # Perform traceroute and save number of hops
                nbHops = Traceroute(dstInfos, proto, id, pktConfig, True, saveDirPath=path).launch()
                nbHopsByProto[proto] = nbHops
            else:
                nbHops = Traceroute(dstInfos, proto, id, pktConfig, False, maxHops=maxHops[proto], saveDirPath=path).launch()
        else:
            nbHops = Traceroute(dstInfos, proto, id, pktConfig, False, maxHops=config.DEFAULT_MAXHOP, saveDirPath=path).launch()
            nbHopsByProto[proto] = config.DEFAULT_MAXHOP

    return nbHopsByProto

"""
Perform the tests sequence between a source and a destination
"""
def source_destination_test(name, dstInfos, expToSkip):
    if config.VERBOSE:
        print('------------ To {} ------------'.format(name))
    else:
        print('[INFO] Probing {}'.format(name))

    # Perform a first traceroute towards the destination
    if config.VERBOSE: print('[INFO] Starting first traceroute (without EH)')
    nbHopsByProto = traceroute_to_destination('StartTrace', name, dstInfos)
    if config.VERBOSE: print('[INFO] First traceroute (without EH) ended')

    # Wait for next step
    time.sleep(config.WAIT_TRACE)

    # Configure the experiments
    if config.VERBOSE: print('[INFO] Starting performing experiments')
    exps = Experiments(config.TESTS_DIR, name, dstInfos, nbHopsByProto, config.PROTOCOLS, expToSkip, config.EXPS_FILENAME)
    exps.start_experiments()
    if config.VERBOSE: print('[INFO] Experiments completed')

    # Wait for next step
    time.sleep(config.WAIT_TRACE)

    # Perform a second traceroute towards the destination
    if config.VERBOSE: print('[INFO] Starting second traceroute (without EH)')
    traceroute_to_destination('EndTrace', name, dstInfos, nbHopsByProto)
    if config.VERBOSE: print('[INFO] Second traceroute (without EH) ended')
    
"""
Start probing
"""
def main():
    # Check if root permission
    if os.getuid() != 0:
        print("[ERROR] This script required root permissions")
        exit(1)
        
    # Parse arguments
    args = parseArgs()

    # Retrieve all the destination addresses and their corresponding information
    dstAddrs = retrieve_destinations('IPv6_dstAddr.txt')
    
    # Create the directory structure
    setup_dir_structure()

    print('##### START TESTS #####')
    # Perform the experiments to the differents destination in sequential manner
    for name, dstInfos in dstAddrs.items():
        # Start capture
        path = config.TESTS_DIR + '/captures/' + name + '_capture.pcap'
        cpt = Capture(args.i, path)
        cpt.start_capture()

        # Ensure capture start
        time.sleep(config.WAIT_CAPTURE)
    
        # Perform test
        source_destination_test(name, dstInfos, args.s)

        # Stop capture
        cpt.stop_capture()

        # Ensure captur stop
        time.sleep(config.WAIT_CAPTURE)

        if config.VERBOSE: print("\n")
    
    return 0
    
if __name__ == "__main__":
    main()
